import datetime

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Field(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    Unknown: _ClassVar[Field]
    DriveRail: _ClassVar[Field]
    ChargeState: _ClassVar[Field]
    BmsFullchargecomplete: _ClassVar[Field]
    VehicleSpeed: _ClassVar[Field]
    Odometer: _ClassVar[Field]
    PackVoltage: _ClassVar[Field]
    PackCurrent: _ClassVar[Field]
    Soc: _ClassVar[Field]
    DCDCEnable: _ClassVar[Field]
    Gear: _ClassVar[Field]
    IsolationResistance: _ClassVar[Field]
    PedalPosition: _ClassVar[Field]
    BrakePedal: _ClassVar[Field]
    DiStateR: _ClassVar[Field]
    DiHeatsinkTR: _ClassVar[Field]
    DiAxleSpeedR: _ClassVar[Field]
    DiTorquemotor: _ClassVar[Field]
    DiStatorTempR: _ClassVar[Field]
    DiVBatR: _ClassVar[Field]
    DiMotorCurrentR: _ClassVar[Field]
    Location: _ClassVar[Field]
    GpsState: _ClassVar[Field]
    GpsHeading: _ClassVar[Field]
    NumBrickVoltageMax: _ClassVar[Field]
    BrickVoltageMax: _ClassVar[Field]
    NumBrickVoltageMin: _ClassVar[Field]
    BrickVoltageMin: _ClassVar[Field]
    NumModuleTempMax: _ClassVar[Field]
    ModuleTempMax: _ClassVar[Field]
    NumModuleTempMin: _ClassVar[Field]
    ModuleTempMin: _ClassVar[Field]
    RatedRange: _ClassVar[Field]
    Hvil: _ClassVar[Field]
    DCChargingEnergyIn: _ClassVar[Field]
    DCChargingPower: _ClassVar[Field]
    ACChargingEnergyIn: _ClassVar[Field]
    ACChargingPower: _ClassVar[Field]
    ChargeLimitSoc: _ClassVar[Field]
    FastChargerPresent: _ClassVar[Field]
    EstBatteryRange: _ClassVar[Field]
    IdealBatteryRange: _ClassVar[Field]
    BatteryLevel: _ClassVar[Field]
    TimeToFullCharge: _ClassVar[Field]
    ScheduledChargingStartTime: _ClassVar[Field]
    ScheduledChargingPending: _ClassVar[Field]
    ScheduledDepartureTime: _ClassVar[Field]
    PreconditioningEnabled: _ClassVar[Field]
    ScheduledChargingMode: _ClassVar[Field]
    ChargeAmps: _ClassVar[Field]
    ChargeEnableRequest: _ClassVar[Field]
    ChargerPhases: _ClassVar[Field]
    ChargePortColdWeatherMode: _ClassVar[Field]
    ChargeCurrentRequest: _ClassVar[Field]
    ChargeCurrentRequestMax: _ClassVar[Field]
    BatteryHeaterOn: _ClassVar[Field]
    NotEnoughPowerToHeat: _ClassVar[Field]
    SuperchargerSessionTripPlanner: _ClassVar[Field]
    DoorState: _ClassVar[Field]
    Locked: _ClassVar[Field]
    FdWindow: _ClassVar[Field]
    FpWindow: _ClassVar[Field]
    RdWindow: _ClassVar[Field]
    RpWindow: _ClassVar[Field]
    VehicleName: _ClassVar[Field]
    SentryMode: _ClassVar[Field]
    SpeedLimitMode: _ClassVar[Field]
    CurrentLimitMph: _ClassVar[Field]
    Version: _ClassVar[Field]
    TpmsPressureFl: _ClassVar[Field]
    TpmsPressureFr: _ClassVar[Field]
    TpmsPressureRl: _ClassVar[Field]
    TpmsPressureRr: _ClassVar[Field]
    SemitruckTpmsPressureRe1L0: _ClassVar[Field]
    SemitruckTpmsPressureRe1L1: _ClassVar[Field]
    SemitruckTpmsPressureRe1R0: _ClassVar[Field]
    SemitruckTpmsPressureRe1R1: _ClassVar[Field]
    SemitruckTpmsPressureRe2L0: _ClassVar[Field]
    SemitruckTpmsPressureRe2L1: _ClassVar[Field]
    SemitruckTpmsPressureRe2R0: _ClassVar[Field]
    SemitruckTpmsPressureRe2R1: _ClassVar[Field]
    TpmsLastSeenPressureTimeFl: _ClassVar[Field]
    TpmsLastSeenPressureTimeFr: _ClassVar[Field]
    TpmsLastSeenPressureTimeRl: _ClassVar[Field]
    TpmsLastSeenPressureTimeRr: _ClassVar[Field]
    InsideTemp: _ClassVar[Field]
    OutsideTemp: _ClassVar[Field]
    SeatHeaterLeft: _ClassVar[Field]
    SeatHeaterRight: _ClassVar[Field]
    SeatHeaterRearLeft: _ClassVar[Field]
    SeatHeaterRearRight: _ClassVar[Field]
    SeatHeaterRearCenter: _ClassVar[Field]
    AutoSeatClimateLeft: _ClassVar[Field]
    AutoSeatClimateRight: _ClassVar[Field]
    DriverSeatBelt: _ClassVar[Field]
    PassengerSeatBelt: _ClassVar[Field]
    DriverSeatOccupied: _ClassVar[Field]
    SemitruckPassengerSeatFoldPosition: _ClassVar[Field]
    LateralAcceleration: _ClassVar[Field]
    LongitudinalAcceleration: _ClassVar[Field]
    Deprecated_2: _ClassVar[Field]
    CruiseSetSpeed: _ClassVar[Field]
    LifetimeEnergyUsed: _ClassVar[Field]
    LifetimeEnergyUsedDrive: _ClassVar[Field]
    SemitruckTractorParkBrakeStatus: _ClassVar[Field]
    SemitruckTrailerParkBrakeStatus: _ClassVar[Field]
    BrakePedalPos: _ClassVar[Field]
    RouteLastUpdated: _ClassVar[Field]
    RouteLine: _ClassVar[Field]
    MilesToArrival: _ClassVar[Field]
    MinutesToArrival: _ClassVar[Field]
    OriginLocation: _ClassVar[Field]
    DestinationLocation: _ClassVar[Field]
    CarType: _ClassVar[Field]
    Trim: _ClassVar[Field]
    ExteriorColor: _ClassVar[Field]
    RoofColor: _ClassVar[Field]
    ChargePort: _ClassVar[Field]
    ChargePortLatch: _ClassVar[Field]
    Experimental_1: _ClassVar[Field]
    Experimental_2: _ClassVar[Field]
    Experimental_3: _ClassVar[Field]
    Experimental_4: _ClassVar[Field]
    GuestModeEnabled: _ClassVar[Field]
    PinToDriveEnabled: _ClassVar[Field]
    PairedPhoneKeyAndKeyFobQty: _ClassVar[Field]
    CruiseFollowDistance: _ClassVar[Field]
    AutomaticBlindSpotCamera: _ClassVar[Field]
    BlindSpotCollisionWarningChime: _ClassVar[Field]
    SpeedLimitWarning: _ClassVar[Field]
    ForwardCollisionWarning: _ClassVar[Field]
    LaneDepartureAvoidance: _ClassVar[Field]
    EmergencyLaneDepartureAvoidance: _ClassVar[Field]
    AutomaticEmergencyBrakingOff: _ClassVar[Field]
    LifetimeEnergyGainedRegen: _ClassVar[Field]
    DiStateF: _ClassVar[Field]
    DiStateREL: _ClassVar[Field]
    DiStateRER: _ClassVar[Field]
    DiHeatsinkTF: _ClassVar[Field]
    DiHeatsinkTREL: _ClassVar[Field]
    DiHeatsinkTRER: _ClassVar[Field]
    DiAxleSpeedF: _ClassVar[Field]
    DiAxleSpeedREL: _ClassVar[Field]
    DiAxleSpeedRER: _ClassVar[Field]
    DiSlaveTorqueCmd: _ClassVar[Field]
    DiTorqueActualR: _ClassVar[Field]
    DiTorqueActualF: _ClassVar[Field]
    DiTorqueActualREL: _ClassVar[Field]
    DiTorqueActualRER: _ClassVar[Field]
    DiStatorTempF: _ClassVar[Field]
    DiStatorTempREL: _ClassVar[Field]
    DiStatorTempRER: _ClassVar[Field]
    DiVBatF: _ClassVar[Field]
    DiVBatREL: _ClassVar[Field]
    DiVBatRER: _ClassVar[Field]
    DiMotorCurrentF: _ClassVar[Field]
    DiMotorCurrentREL: _ClassVar[Field]
    DiMotorCurrentRER: _ClassVar[Field]
    EnergyRemaining: _ClassVar[Field]
    ServiceMode: _ClassVar[Field]
    BMSState: _ClassVar[Field]
    GuestModeMobileAccessState: _ClassVar[Field]
    Deprecated_1: _ClassVar[Field]
    DestinationName: _ClassVar[Field]
    DiInverterTR: _ClassVar[Field]
    DiInverterTF: _ClassVar[Field]
    DiInverterTREL: _ClassVar[Field]
    DiInverterTRER: _ClassVar[Field]
    Experimental_5: _ClassVar[Field]
    Experimental_6: _ClassVar[Field]
    Experimental_7: _ClassVar[Field]
    Experimental_8: _ClassVar[Field]
    Experimental_9: _ClassVar[Field]
    Experimental_10: _ClassVar[Field]
    Experimental_11: _ClassVar[Field]
    Experimental_12: _ClassVar[Field]
    Experimental_13: _ClassVar[Field]
    Experimental_14: _ClassVar[Field]
    Experimental_15: _ClassVar[Field]
    DetailedChargeState: _ClassVar[Field]
    CabinOverheatProtectionMode: _ClassVar[Field]
    CabinOverheatProtectionTemperatureLimit: _ClassVar[Field]
    CenterDisplay: _ClassVar[Field]
    ChargePortDoorOpen: _ClassVar[Field]
    ChargerVoltage: _ClassVar[Field]
    ChargingCableType: _ClassVar[Field]
    ClimateKeeperMode: _ClassVar[Field]
    DefrostForPreconditioning: _ClassVar[Field]
    DefrostMode: _ClassVar[Field]
    EfficiencyPackage: _ClassVar[Field]
    EstimatedHoursToChargeTermination: _ClassVar[Field]
    EuropeVehicle: _ClassVar[Field]
    ExpectedEnergyPercentAtTripArrival: _ClassVar[Field]
    FastChargerType: _ClassVar[Field]
    HomelinkDeviceCount: _ClassVar[Field]
    HomelinkNearby: _ClassVar[Field]
    HvacACEnabled: _ClassVar[Field]
    HvacAutoMode: _ClassVar[Field]
    HvacFanSpeed: _ClassVar[Field]
    HvacFanStatus: _ClassVar[Field]
    HvacLeftTemperatureRequest: _ClassVar[Field]
    HvacPower: _ClassVar[Field]
    HvacRightTemperatureRequest: _ClassVar[Field]
    HvacSteeringWheelHeatAuto: _ClassVar[Field]
    HvacSteeringWheelHeatLevel: _ClassVar[Field]
    OffroadLightbarPresent: _ClassVar[Field]
    PowershareHoursLeft: _ClassVar[Field]
    PowershareInstantaneousPowerKW: _ClassVar[Field]
    PowershareStatus: _ClassVar[Field]
    PowershareStopReason: _ClassVar[Field]
    PowershareType: _ClassVar[Field]
    RearDisplayHvacEnabled: _ClassVar[Field]
    RearSeatHeaters: _ClassVar[Field]
    RemoteStartEnabled: _ClassVar[Field]
    RightHandDrive: _ClassVar[Field]
    RouteTrafficMinutesDelay: _ClassVar[Field]
    SoftwareUpdateDownloadPercentComplete: _ClassVar[Field]
    SoftwareUpdateExpectedDurationMinutes: _ClassVar[Field]
    SoftwareUpdateInstallationPercentComplete: _ClassVar[Field]
    SoftwareUpdateScheduledStartTime: _ClassVar[Field]
    SoftwareUpdateVersion: _ClassVar[Field]
    TonneauOpenPercent: _ClassVar[Field]
    TonneauPosition: _ClassVar[Field]
    TonneauTentMode: _ClassVar[Field]
    TpmsHardWarnings: _ClassVar[Field]
    TpmsSoftWarnings: _ClassVar[Field]
    ValetModeEnabled: _ClassVar[Field]
    WheelType: _ClassVar[Field]
    WiperHeatEnabled: _ClassVar[Field]
    LocatedAtHome: _ClassVar[Field]
    LocatedAtWork: _ClassVar[Field]
    LocatedAtFavorite: _ClassVar[Field]
    SettingDistanceUnit: _ClassVar[Field]
    SettingTemperatureUnit: _ClassVar[Field]
    Setting24HourTime: _ClassVar[Field]
    SettingTirePressureUnit: _ClassVar[Field]
    SettingChargeUnit: _ClassVar[Field]
    ClimateSeatCoolingFrontLeft: _ClassVar[Field]
    ClimateSeatCoolingFrontRight: _ClassVar[Field]
    LightsHazardsActive: _ClassVar[Field]
    LightsTurnSignal: _ClassVar[Field]
    LightsHighBeams: _ClassVar[Field]
    MediaPlaybackStatus: _ClassVar[Field]
    MediaPlaybackSource: _ClassVar[Field]
    MediaAudioVolume: _ClassVar[Field]
    MediaNowPlayingDuration: _ClassVar[Field]
    MediaNowPlayingElapsed: _ClassVar[Field]
    MediaNowPlayingArtist: _ClassVar[Field]
    MediaNowPlayingTitle: _ClassVar[Field]
    MediaNowPlayingAlbum: _ClassVar[Field]
    MediaNowPlayingStation: _ClassVar[Field]
    MediaAudioVolumeIncrement: _ClassVar[Field]
    MediaAudioVolumeMax: _ClassVar[Field]
    SunroofInstalled: _ClassVar[Field]
    SeatVentEnabled: _ClassVar[Field]
    RearDefrostEnabled: _ClassVar[Field]
    ChargeRateMilePerHour: _ClassVar[Field]
    Deprecated_3: _ClassVar[Field]
    MilesSinceReset: _ClassVar[Field]
    SelfDrivingMilesSinceReset: _ClassVar[Field]

class ChargingState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ChargeStateUnknown: _ClassVar[ChargingState]
    ChargeStateDisconnected: _ClassVar[ChargingState]
    ChargeStateNoPower: _ClassVar[ChargingState]
    ChargeStateStarting: _ClassVar[ChargingState]
    ChargeStateCharging: _ClassVar[ChargingState]
    ChargeStateComplete: _ClassVar[ChargingState]
    ChargeStateStopped: _ClassVar[ChargingState]

class DetailedChargeStateValue(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DetailedChargeStateUnknown: _ClassVar[DetailedChargeStateValue]
    DetailedChargeStateDisconnected: _ClassVar[DetailedChargeStateValue]
    DetailedChargeStateNoPower: _ClassVar[DetailedChargeStateValue]
    DetailedChargeStateStarting: _ClassVar[DetailedChargeStateValue]
    DetailedChargeStateCharging: _ClassVar[DetailedChargeStateValue]
    DetailedChargeStateComplete: _ClassVar[DetailedChargeStateValue]
    DetailedChargeStateStopped: _ClassVar[DetailedChargeStateValue]

class ShiftState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ShiftStateUnknown: _ClassVar[ShiftState]
    ShiftStateInvalid: _ClassVar[ShiftState]
    ShiftStateP: _ClassVar[ShiftState]
    ShiftStateR: _ClassVar[ShiftState]
    ShiftStateN: _ClassVar[ShiftState]
    ShiftStateD: _ClassVar[ShiftState]
    ShiftStateSNA: _ClassVar[ShiftState]

class FollowDistance(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FollowDistanceUnknown: _ClassVar[FollowDistance]
    FollowDistance1: _ClassVar[FollowDistance]
    FollowDistance2: _ClassVar[FollowDistance]
    FollowDistance3: _ClassVar[FollowDistance]
    FollowDistance4: _ClassVar[FollowDistance]
    FollowDistance5: _ClassVar[FollowDistance]
    FollowDistance6: _ClassVar[FollowDistance]
    FollowDistance7: _ClassVar[FollowDistance]

class ForwardCollisionSensitivity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ForwardCollisionSensitivityUnknown: _ClassVar[ForwardCollisionSensitivity]
    ForwardCollisionSensitivityOff: _ClassVar[ForwardCollisionSensitivity]
    ForwardCollisionSensitivityLate: _ClassVar[ForwardCollisionSensitivity]
    ForwardCollisionSensitivityAverage: _ClassVar[ForwardCollisionSensitivity]
    ForwardCollisionSensitivityEarly: _ClassVar[ForwardCollisionSensitivity]

class GuestModeMobileAccess(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    GuestModeMobileAccessUnknown: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessInit: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessNotAuthenticated: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAuthenticated: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedDriving: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedUsingRemoteStart: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedUsingBLEKeys: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedValetMode: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedGuestModeOff: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedDriveAuthTimeExceeded: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedNoDataReceived: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessRequestingFromMothership: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessRequestingFromAuthD: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedFetchFailed: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessAbortedBadDataReceived: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessShowingQRCode: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessSwipedAway: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessDismissedQRCodeExpired: _ClassVar[GuestModeMobileAccess]
    GuestModeMobileAccessSucceededPairedNewBLEKey: _ClassVar[GuestModeMobileAccess]

class LaneAssistLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    LaneAssistLevelUnknown: _ClassVar[LaneAssistLevel]
    LaneAssistLevelNone: _ClassVar[LaneAssistLevel]
    LaneAssistLevelWarning: _ClassVar[LaneAssistLevel]
    LaneAssistLevelAssist: _ClassVar[LaneAssistLevel]

class ScheduledChargingModeValue(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ScheduledChargingModeUnknown: _ClassVar[ScheduledChargingModeValue]
    ScheduledChargingModeOff: _ClassVar[ScheduledChargingModeValue]
    ScheduledChargingModeStartAt: _ClassVar[ScheduledChargingModeValue]
    ScheduledChargingModeDepartBy: _ClassVar[ScheduledChargingModeValue]

class SentryModeState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SentryModeStateUnknown: _ClassVar[SentryModeState]
    SentryModeStateOff: _ClassVar[SentryModeState]
    SentryModeStateIdle: _ClassVar[SentryModeState]
    SentryModeStateArmed: _ClassVar[SentryModeState]
    SentryModeStateAware: _ClassVar[SentryModeState]
    SentryModeStatePanic: _ClassVar[SentryModeState]
    SentryModeStateQuiet: _ClassVar[SentryModeState]

class SpeedAssistLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SpeedAssistLevelUnknown: _ClassVar[SpeedAssistLevel]
    SpeedAssistLevelNone: _ClassVar[SpeedAssistLevel]
    SpeedAssistLevelDisplay: _ClassVar[SpeedAssistLevel]
    SpeedAssistLevelChime: _ClassVar[SpeedAssistLevel]

class BMSStateValue(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BMSStateUnknown: _ClassVar[BMSStateValue]
    BMSStateStandby: _ClassVar[BMSStateValue]
    BMSStateDrive: _ClassVar[BMSStateValue]
    BMSStateSupport: _ClassVar[BMSStateValue]
    BMSStateCharge: _ClassVar[BMSStateValue]
    BMSStateFEIM: _ClassVar[BMSStateValue]
    BMSStateClearFault: _ClassVar[BMSStateValue]
    BMSStateFault: _ClassVar[BMSStateValue]
    BMSStateWeld: _ClassVar[BMSStateValue]
    BMSStateTest: _ClassVar[BMSStateValue]
    BMSStateSNA: _ClassVar[BMSStateValue]

class BuckleStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BuckleStatusUnknown: _ClassVar[BuckleStatus]
    BuckleStatusUnlatched: _ClassVar[BuckleStatus]
    BuckleStatusLatched: _ClassVar[BuckleStatus]
    BuckleStatusFaulted: _ClassVar[BuckleStatus]

class CarTypeValue(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CarTypeUnknown: _ClassVar[CarTypeValue]
    CarTypeModelS: _ClassVar[CarTypeValue]
    CarTypeModelX: _ClassVar[CarTypeValue]
    CarTypeModel3: _ClassVar[CarTypeValue]
    CarTypeModelY: _ClassVar[CarTypeValue]
    CarTypeSemiTruck: _ClassVar[CarTypeValue]
    CarTypeCybertruck: _ClassVar[CarTypeValue]

class ChargePortValue(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ChargePortUnknown: _ClassVar[ChargePortValue]
    ChargePortUS: _ClassVar[ChargePortValue]
    ChargePortEU: _ClassVar[ChargePortValue]
    ChargePortGB: _ClassVar[ChargePortValue]
    ChargePortCCS: _ClassVar[ChargePortValue]

class ChargePortLatchValue(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ChargePortLatchUnknown: _ClassVar[ChargePortLatchValue]
    ChargePortLatchSNA: _ClassVar[ChargePortLatchValue]
    ChargePortLatchDisengaged: _ClassVar[ChargePortLatchValue]
    ChargePortLatchEngaged: _ClassVar[ChargePortLatchValue]
    ChargePortLatchBlocking: _ClassVar[ChargePortLatchValue]

class DriveInverterState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DriveInverterStateUnknown: _ClassVar[DriveInverterState]
    DriveInverterStateUnavailable: _ClassVar[DriveInverterState]
    DriveInverterStateStandby: _ClassVar[DriveInverterState]
    DriveInverterStateFault: _ClassVar[DriveInverterState]
    DriveInverterStateAbort: _ClassVar[DriveInverterState]
    DriveInverterStateEnable: _ClassVar[DriveInverterState]

class HvilStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    HvilStatusUnknown: _ClassVar[HvilStatus]
    HvilStatusFault: _ClassVar[HvilStatus]
    HvilStatusOK: _ClassVar[HvilStatus]

class WindowState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    WindowStateUnknown: _ClassVar[WindowState]
    WindowStateClosed: _ClassVar[WindowState]
    WindowStatePartiallyOpen: _ClassVar[WindowState]
    WindowStateOpened: _ClassVar[WindowState]

class SeatFoldPosition(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SeatFoldPositionUnknown: _ClassVar[SeatFoldPosition]
    SeatFoldPositionSNA: _ClassVar[SeatFoldPosition]
    SeatFoldPositionFaulted: _ClassVar[SeatFoldPosition]
    SeatFoldPositionNotConfigured: _ClassVar[SeatFoldPosition]
    SeatFoldPositionFolded: _ClassVar[SeatFoldPosition]
    SeatFoldPositionUnfolded: _ClassVar[SeatFoldPosition]

class TractorAirStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TractorAirStatusUnknown: _ClassVar[TractorAirStatus]
    TractorAirStatusNotAvailable: _ClassVar[TractorAirStatus]
    TractorAirStatusError: _ClassVar[TractorAirStatus]
    TractorAirStatusCharged: _ClassVar[TractorAirStatus]
    TractorAirStatusBuildingPressureIntermediate: _ClassVar[TractorAirStatus]
    TractorAirStatusExhaustingPressureIntermediate: _ClassVar[TractorAirStatus]
    TractorAirStatusExhausted: _ClassVar[TractorAirStatus]

class TrailerAirStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TrailerAirStatusUnknown: _ClassVar[TrailerAirStatus]
    TrailerAirStatusSNA: _ClassVar[TrailerAirStatus]
    TrailerAirStatusInvalid: _ClassVar[TrailerAirStatus]
    TrailerAirStatusBobtailMode: _ClassVar[TrailerAirStatus]
    TrailerAirStatusCharged: _ClassVar[TrailerAirStatus]
    TrailerAirStatusBuildingPressureIntermediate: _ClassVar[TrailerAirStatus]
    TrailerAirStatusExhaustingPressureIntermediate: _ClassVar[TrailerAirStatus]
    TrailerAirStatusExhausted: _ClassVar[TrailerAirStatus]

class HvacAutoModeState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    HvacAutoModeStateUnknown: _ClassVar[HvacAutoModeState]
    HvacAutoModeStateOn: _ClassVar[HvacAutoModeState]
    HvacAutoModeStateOverride: _ClassVar[HvacAutoModeState]

class CabinOverheatProtectionModeState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CabinOverheatProtectionModeStateUnknown: _ClassVar[CabinOverheatProtectionModeState]
    CabinOverheatProtectionModeStateOff: _ClassVar[CabinOverheatProtectionModeState]
    CabinOverheatProtectionModeStateOn: _ClassVar[CabinOverheatProtectionModeState]
    CabinOverheatProtectionModeStateFanOnly: _ClassVar[CabinOverheatProtectionModeState]

class ClimateOverheatProtectionTempLimit(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ClimateOverheatProtectionTempLimitUnknown: _ClassVar[ClimateOverheatProtectionTempLimit]
    ClimateOverheatProtectionTempLimitHigh: _ClassVar[ClimateOverheatProtectionTempLimit]
    ClimateOverheatProtectionTempLimitMedium: _ClassVar[ClimateOverheatProtectionTempLimit]
    ClimateOverheatProtectionTempLimitLow: _ClassVar[ClimateOverheatProtectionTempLimit]

class DefrostModeState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DefrostModeStateUnknown: _ClassVar[DefrostModeState]
    DefrostModeStateOff: _ClassVar[DefrostModeState]
    DefrostModeStateNormal: _ClassVar[DefrostModeState]
    DefrostModeStateMax: _ClassVar[DefrostModeState]
    DefrostModeStateAutoDefog: _ClassVar[DefrostModeState]

class ClimateKeeperModeState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ClimateKeeperModeStateUnknown: _ClassVar[ClimateKeeperModeState]
    ClimateKeeperModeStateOff: _ClassVar[ClimateKeeperModeState]
    ClimateKeeperModeStateOn: _ClassVar[ClimateKeeperModeState]
    ClimateKeeperModeStateDog: _ClassVar[ClimateKeeperModeState]
    ClimateKeeperModeStateParty: _ClassVar[ClimateKeeperModeState]

class HvacPowerState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    HvacPowerStateUnknown: _ClassVar[HvacPowerState]
    HvacPowerStateOff: _ClassVar[HvacPowerState]
    HvacPowerStateOn: _ClassVar[HvacPowerState]
    HvacPowerStatePrecondition: _ClassVar[HvacPowerState]
    HvacPowerStateOverheatProtect: _ClassVar[HvacPowerState]

class FastCharger(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FastChargerUnknown: _ClassVar[FastCharger]
    FastChargerSupercharger: _ClassVar[FastCharger]
    FastChargerCHAdeMO: _ClassVar[FastCharger]
    FastChargerGB: _ClassVar[FastCharger]
    FastChargerACSingleWireCAN: _ClassVar[FastCharger]
    FastChargerCombo: _ClassVar[FastCharger]
    FastChargerMCSingleWireCAN: _ClassVar[FastCharger]
    FastChargerOther: _ClassVar[FastCharger]
    FastChargerSNA: _ClassVar[FastCharger]

class CableType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CableTypeUnknown: _ClassVar[CableType]
    CableTypeIEC: _ClassVar[CableType]
    CableTypeSAE: _ClassVar[CableType]
    CableTypeGB_AC: _ClassVar[CableType]
    CableTypeGB_DC: _ClassVar[CableType]
    CableTypeSNA: _ClassVar[CableType]

class TonneauTentModeState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TonneauTentModeStateUnknown: _ClassVar[TonneauTentModeState]
    TonneauTentModeStateInactive: _ClassVar[TonneauTentModeState]
    TonneauTentModeStateMoving: _ClassVar[TonneauTentModeState]
    TonneauTentModeStateFailed: _ClassVar[TonneauTentModeState]
    TonneauTentModeStateActive: _ClassVar[TonneauTentModeState]

class TonneauPositionState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TonneauPositionStateUnknown: _ClassVar[TonneauPositionState]
    TonneauPositionStateInvalid: _ClassVar[TonneauPositionState]
    TonneauPositionStateClosed: _ClassVar[TonneauPositionState]
    TonneauPositionStatePartiallyOpen: _ClassVar[TonneauPositionState]
    TonneauPositionStateFullyOpen: _ClassVar[TonneauPositionState]

class PowershareState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PowershareStateUnknown: _ClassVar[PowershareState]
    PowershareStateInactive: _ClassVar[PowershareState]
    PowershareStateHandshaking: _ClassVar[PowershareState]
    PowershareStateInit: _ClassVar[PowershareState]
    PowershareStateEnabled: _ClassVar[PowershareState]
    PowershareStateEnabledReconnectingSoon: _ClassVar[PowershareState]
    PowershareStateStopped: _ClassVar[PowershareState]

class PowershareStopReasonStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PowershareStopReasonStatusUnknown: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusNone: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusSOCTooLow: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusRetry: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusFault: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusUser: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusReconnecting: _ClassVar[PowershareStopReasonStatus]
    PowershareStopReasonStatusAuthentication: _ClassVar[PowershareStopReasonStatus]

class PowershareTypeStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PowershareTypeStatusUnknown: _ClassVar[PowershareTypeStatus]
    PowershareTypeStatusNone: _ClassVar[PowershareTypeStatus]
    PowershareTypeStatusLoad: _ClassVar[PowershareTypeStatus]
    PowershareTypeStatusHome: _ClassVar[PowershareTypeStatus]

class DisplayState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DisplayStateUnknown: _ClassVar[DisplayState]
    DisplayStateOff: _ClassVar[DisplayState]
    DisplayStateDim: _ClassVar[DisplayState]
    DisplayStateAccessory: _ClassVar[DisplayState]
    DisplayStateOn: _ClassVar[DisplayState]
    DisplayStateDriving: _ClassVar[DisplayState]
    DisplayStateCharging: _ClassVar[DisplayState]
    DisplayStateLock: _ClassVar[DisplayState]
    DisplayStateSentry: _ClassVar[DisplayState]
    DisplayStateDog: _ClassVar[DisplayState]
    DisplayStateEntertainment: _ClassVar[DisplayState]

class DistanceUnit(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DistanceUnitUnknown: _ClassVar[DistanceUnit]
    DistanceUnitMiles: _ClassVar[DistanceUnit]
    DistanceUnitKilometers: _ClassVar[DistanceUnit]

class TemperatureUnit(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TemperatureUnitUnknown: _ClassVar[TemperatureUnit]
    TemperatureUnitFahrenheit: _ClassVar[TemperatureUnit]
    TemperatureUnitCelsius: _ClassVar[TemperatureUnit]

class PressureUnit(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PressureUnitUnknown: _ClassVar[PressureUnit]
    PressureUnitPsi: _ClassVar[PressureUnit]
    PressureUnitBar: _ClassVar[PressureUnit]

class ChargeUnitPreference(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ChargeUnitUnknown: _ClassVar[ChargeUnitPreference]
    ChargeUnitDistance: _ClassVar[ChargeUnitPreference]
    ChargeUnitPercent: _ClassVar[ChargeUnitPreference]

class SunroofInstalledState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SunroofInstalledStateUnknown: _ClassVar[SunroofInstalledState]
    SunroofInstalledStateNotInstalled: _ClassVar[SunroofInstalledState]
    SunroofInstalledStateGen1Installed: _ClassVar[SunroofInstalledState]
    SunroofInstalledStateGen2Installed: _ClassVar[SunroofInstalledState]

class TurnSignalState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TurnSignalStateUnknown: _ClassVar[TurnSignalState]
    TurnSignalStateOff: _ClassVar[TurnSignalState]
    TurnSignalStateLeft: _ClassVar[TurnSignalState]
    TurnSignalStateRight: _ClassVar[TurnSignalState]
    TurnSignalStateBoth: _ClassVar[TurnSignalState]

class MediaStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    MediaStatusUnknown: _ClassVar[MediaStatus]
    MediaStatusStopped: _ClassVar[MediaStatus]
    MediaStatusPlaying: _ClassVar[MediaStatus]
    MediaStatusPaused: _ClassVar[MediaStatus]
Unknown: Field
DriveRail: Field
ChargeState: Field
BmsFullchargecomplete: Field
VehicleSpeed: Field
Odometer: Field
PackVoltage: Field
PackCurrent: Field
Soc: Field
DCDCEnable: Field
Gear: Field
IsolationResistance: Field
PedalPosition: Field
BrakePedal: Field
DiStateR: Field
DiHeatsinkTR: Field
DiAxleSpeedR: Field
DiTorquemotor: Field
DiStatorTempR: Field
DiVBatR: Field
DiMotorCurrentR: Field
Location: Field
GpsState: Field
GpsHeading: Field
NumBrickVoltageMax: Field
BrickVoltageMax: Field
NumBrickVoltageMin: Field
BrickVoltageMin: Field
NumModuleTempMax: Field
ModuleTempMax: Field
NumModuleTempMin: Field
ModuleTempMin: Field
RatedRange: Field
Hvil: Field
DCChargingEnergyIn: Field
DCChargingPower: Field
ACChargingEnergyIn: Field
ACChargingPower: Field
ChargeLimitSoc: Field
FastChargerPresent: Field
EstBatteryRange: Field
IdealBatteryRange: Field
BatteryLevel: Field
TimeToFullCharge: Field
ScheduledChargingStartTime: Field
ScheduledChargingPending: Field
ScheduledDepartureTime: Field
PreconditioningEnabled: Field
ScheduledChargingMode: Field
ChargeAmps: Field
ChargeEnableRequest: Field
ChargerPhases: Field
ChargePortColdWeatherMode: Field
ChargeCurrentRequest: Field
ChargeCurrentRequestMax: Field
BatteryHeaterOn: Field
NotEnoughPowerToHeat: Field
SuperchargerSessionTripPlanner: Field
DoorState: Field
Locked: Field
FdWindow: Field
FpWindow: Field
RdWindow: Field
RpWindow: Field
VehicleName: Field
SentryMode: Field
SpeedLimitMode: Field
CurrentLimitMph: Field
Version: Field
TpmsPressureFl: Field
TpmsPressureFr: Field
TpmsPressureRl: Field
TpmsPressureRr: Field
SemitruckTpmsPressureRe1L0: Field
SemitruckTpmsPressureRe1L1: Field
SemitruckTpmsPressureRe1R0: Field
SemitruckTpmsPressureRe1R1: Field
SemitruckTpmsPressureRe2L0: Field
SemitruckTpmsPressureRe2L1: Field
SemitruckTpmsPressureRe2R0: Field
SemitruckTpmsPressureRe2R1: Field
TpmsLastSeenPressureTimeFl: Field
TpmsLastSeenPressureTimeFr: Field
TpmsLastSeenPressureTimeRl: Field
TpmsLastSeenPressureTimeRr: Field
InsideTemp: Field
OutsideTemp: Field
SeatHeaterLeft: Field
SeatHeaterRight: Field
SeatHeaterRearLeft: Field
SeatHeaterRearRight: Field
SeatHeaterRearCenter: Field
AutoSeatClimateLeft: Field
AutoSeatClimateRight: Field
DriverSeatBelt: Field
PassengerSeatBelt: Field
DriverSeatOccupied: Field
SemitruckPassengerSeatFoldPosition: Field
LateralAcceleration: Field
LongitudinalAcceleration: Field
Deprecated_2: Field
CruiseSetSpeed: Field
LifetimeEnergyUsed: Field
LifetimeEnergyUsedDrive: Field
SemitruckTractorParkBrakeStatus: Field
SemitruckTrailerParkBrakeStatus: Field
BrakePedalPos: Field
RouteLastUpdated: Field
RouteLine: Field
MilesToArrival: Field
MinutesToArrival: Field
OriginLocation: Field
DestinationLocation: Field
CarType: Field
Trim: Field
ExteriorColor: Field
RoofColor: Field
ChargePort: Field
ChargePortLatch: Field
Experimental_1: Field
Experimental_2: Field
Experimental_3: Field
Experimental_4: Field
GuestModeEnabled: Field
PinToDriveEnabled: Field
PairedPhoneKeyAndKeyFobQty: Field
CruiseFollowDistance: Field
AutomaticBlindSpotCamera: Field
BlindSpotCollisionWarningChime: Field
SpeedLimitWarning: Field
ForwardCollisionWarning: Field
LaneDepartureAvoidance: Field
EmergencyLaneDepartureAvoidance: Field
AutomaticEmergencyBrakingOff: Field
LifetimeEnergyGainedRegen: Field
DiStateF: Field
DiStateREL: Field
DiStateRER: Field
DiHeatsinkTF: Field
DiHeatsinkTREL: Field
DiHeatsinkTRER: Field
DiAxleSpeedF: Field
DiAxleSpeedREL: Field
DiAxleSpeedRER: Field
DiSlaveTorqueCmd: Field
DiTorqueActualR: Field
DiTorqueActualF: Field
DiTorqueActualREL: Field
DiTorqueActualRER: Field
DiStatorTempF: Field
DiStatorTempREL: Field
DiStatorTempRER: Field
DiVBatF: Field
DiVBatREL: Field
DiVBatRER: Field
DiMotorCurrentF: Field
DiMotorCurrentREL: Field
DiMotorCurrentRER: Field
EnergyRemaining: Field
ServiceMode: Field
BMSState: Field
GuestModeMobileAccessState: Field
Deprecated_1: Field
DestinationName: Field
DiInverterTR: Field
DiInverterTF: Field
DiInverterTREL: Field
DiInverterTRER: Field
Experimental_5: Field
Experimental_6: Field
Experimental_7: Field
Experimental_8: Field
Experimental_9: Field
Experimental_10: Field
Experimental_11: Field
Experimental_12: Field
Experimental_13: Field
Experimental_14: Field
Experimental_15: Field
DetailedChargeState: Field
CabinOverheatProtectionMode: Field
CabinOverheatProtectionTemperatureLimit: Field
CenterDisplay: Field
ChargePortDoorOpen: Field
ChargerVoltage: Field
ChargingCableType: Field
ClimateKeeperMode: Field
DefrostForPreconditioning: Field
DefrostMode: Field
EfficiencyPackage: Field
EstimatedHoursToChargeTermination: Field
EuropeVehicle: Field
ExpectedEnergyPercentAtTripArrival: Field
FastChargerType: Field
HomelinkDeviceCount: Field
HomelinkNearby: Field
HvacACEnabled: Field
HvacAutoMode: Field
HvacFanSpeed: Field
HvacFanStatus: Field
HvacLeftTemperatureRequest: Field
HvacPower: Field
HvacRightTemperatureRequest: Field
HvacSteeringWheelHeatAuto: Field
HvacSteeringWheelHeatLevel: Field
OffroadLightbarPresent: Field
PowershareHoursLeft: Field
PowershareInstantaneousPowerKW: Field
PowershareStatus: Field
PowershareStopReason: Field
PowershareType: Field
RearDisplayHvacEnabled: Field
RearSeatHeaters: Field
RemoteStartEnabled: Field
RightHandDrive: Field
RouteTrafficMinutesDelay: Field
SoftwareUpdateDownloadPercentComplete: Field
SoftwareUpdateExpectedDurationMinutes: Field
SoftwareUpdateInstallationPercentComplete: Field
SoftwareUpdateScheduledStartTime: Field
SoftwareUpdateVersion: Field
TonneauOpenPercent: Field
TonneauPosition: Field
TonneauTentMode: Field
TpmsHardWarnings: Field
TpmsSoftWarnings: Field
ValetModeEnabled: Field
WheelType: Field
WiperHeatEnabled: Field
LocatedAtHome: Field
LocatedAtWork: Field
LocatedAtFavorite: Field
SettingDistanceUnit: Field
SettingTemperatureUnit: Field
Setting24HourTime: Field
SettingTirePressureUnit: Field
SettingChargeUnit: Field
ClimateSeatCoolingFrontLeft: Field
ClimateSeatCoolingFrontRight: Field
LightsHazardsActive: Field
LightsTurnSignal: Field
LightsHighBeams: Field
MediaPlaybackStatus: Field
MediaPlaybackSource: Field
MediaAudioVolume: Field
MediaNowPlayingDuration: Field
MediaNowPlayingElapsed: Field
MediaNowPlayingArtist: Field
MediaNowPlayingTitle: Field
MediaNowPlayingAlbum: Field
MediaNowPlayingStation: Field
MediaAudioVolumeIncrement: Field
MediaAudioVolumeMax: Field
SunroofInstalled: Field
SeatVentEnabled: Field
RearDefrostEnabled: Field
ChargeRateMilePerHour: Field
Deprecated_3: Field
MilesSinceReset: Field
SelfDrivingMilesSinceReset: Field
ChargeStateUnknown: ChargingState
ChargeStateDisconnected: ChargingState
ChargeStateNoPower: ChargingState
ChargeStateStarting: ChargingState
ChargeStateCharging: ChargingState
ChargeStateComplete: ChargingState
ChargeStateStopped: ChargingState
DetailedChargeStateUnknown: DetailedChargeStateValue
DetailedChargeStateDisconnected: DetailedChargeStateValue
DetailedChargeStateNoPower: DetailedChargeStateValue
DetailedChargeStateStarting: DetailedChargeStateValue
DetailedChargeStateCharging: DetailedChargeStateValue
DetailedChargeStateComplete: DetailedChargeStateValue
DetailedChargeStateStopped: DetailedChargeStateValue
ShiftStateUnknown: ShiftState
ShiftStateInvalid: ShiftState
ShiftStateP: ShiftState
ShiftStateR: ShiftState
ShiftStateN: ShiftState
ShiftStateD: ShiftState
ShiftStateSNA: ShiftState
FollowDistanceUnknown: FollowDistance
FollowDistance1: FollowDistance
FollowDistance2: FollowDistance
FollowDistance3: FollowDistance
FollowDistance4: FollowDistance
FollowDistance5: FollowDistance
FollowDistance6: FollowDistance
FollowDistance7: FollowDistance
ForwardCollisionSensitivityUnknown: ForwardCollisionSensitivity
ForwardCollisionSensitivityOff: ForwardCollisionSensitivity
ForwardCollisionSensitivityLate: ForwardCollisionSensitivity
ForwardCollisionSensitivityAverage: ForwardCollisionSensitivity
ForwardCollisionSensitivityEarly: ForwardCollisionSensitivity
GuestModeMobileAccessUnknown: GuestModeMobileAccess
GuestModeMobileAccessInit: GuestModeMobileAccess
GuestModeMobileAccessNotAuthenticated: GuestModeMobileAccess
GuestModeMobileAccessAuthenticated: GuestModeMobileAccess
GuestModeMobileAccessAbortedDriving: GuestModeMobileAccess
GuestModeMobileAccessAbortedUsingRemoteStart: GuestModeMobileAccess
GuestModeMobileAccessAbortedUsingBLEKeys: GuestModeMobileAccess
GuestModeMobileAccessAbortedValetMode: GuestModeMobileAccess
GuestModeMobileAccessAbortedGuestModeOff: GuestModeMobileAccess
GuestModeMobileAccessAbortedDriveAuthTimeExceeded: GuestModeMobileAccess
GuestModeMobileAccessAbortedNoDataReceived: GuestModeMobileAccess
GuestModeMobileAccessRequestingFromMothership: GuestModeMobileAccess
GuestModeMobileAccessRequestingFromAuthD: GuestModeMobileAccess
GuestModeMobileAccessAbortedFetchFailed: GuestModeMobileAccess
GuestModeMobileAccessAbortedBadDataReceived: GuestModeMobileAccess
GuestModeMobileAccessShowingQRCode: GuestModeMobileAccess
GuestModeMobileAccessSwipedAway: GuestModeMobileAccess
GuestModeMobileAccessDismissedQRCodeExpired: GuestModeMobileAccess
GuestModeMobileAccessSucceededPairedNewBLEKey: GuestModeMobileAccess
LaneAssistLevelUnknown: LaneAssistLevel
LaneAssistLevelNone: LaneAssistLevel
LaneAssistLevelWarning: LaneAssistLevel
LaneAssistLevelAssist: LaneAssistLevel
ScheduledChargingModeUnknown: ScheduledChargingModeValue
ScheduledChargingModeOff: ScheduledChargingModeValue
ScheduledChargingModeStartAt: ScheduledChargingModeValue
ScheduledChargingModeDepartBy: ScheduledChargingModeValue
SentryModeStateUnknown: SentryModeState
SentryModeStateOff: SentryModeState
SentryModeStateIdle: SentryModeState
SentryModeStateArmed: SentryModeState
SentryModeStateAware: SentryModeState
SentryModeStatePanic: SentryModeState
SentryModeStateQuiet: SentryModeState
SpeedAssistLevelUnknown: SpeedAssistLevel
SpeedAssistLevelNone: SpeedAssistLevel
SpeedAssistLevelDisplay: SpeedAssistLevel
SpeedAssistLevelChime: SpeedAssistLevel
BMSStateUnknown: BMSStateValue
BMSStateStandby: BMSStateValue
BMSStateDrive: BMSStateValue
BMSStateSupport: BMSStateValue
BMSStateCharge: BMSStateValue
BMSStateFEIM: BMSStateValue
BMSStateClearFault: BMSStateValue
BMSStateFault: BMSStateValue
BMSStateWeld: BMSStateValue
BMSStateTest: BMSStateValue
BMSStateSNA: BMSStateValue
BuckleStatusUnknown: BuckleStatus
BuckleStatusUnlatched: BuckleStatus
BuckleStatusLatched: BuckleStatus
BuckleStatusFaulted: BuckleStatus
CarTypeUnknown: CarTypeValue
CarTypeModelS: CarTypeValue
CarTypeModelX: CarTypeValue
CarTypeModel3: CarTypeValue
CarTypeModelY: CarTypeValue
CarTypeSemiTruck: CarTypeValue
CarTypeCybertruck: CarTypeValue
ChargePortUnknown: ChargePortValue
ChargePortUS: ChargePortValue
ChargePortEU: ChargePortValue
ChargePortGB: ChargePortValue
ChargePortCCS: ChargePortValue
ChargePortLatchUnknown: ChargePortLatchValue
ChargePortLatchSNA: ChargePortLatchValue
ChargePortLatchDisengaged: ChargePortLatchValue
ChargePortLatchEngaged: ChargePortLatchValue
ChargePortLatchBlocking: ChargePortLatchValue
DriveInverterStateUnknown: DriveInverterState
DriveInverterStateUnavailable: DriveInverterState
DriveInverterStateStandby: DriveInverterState
DriveInverterStateFault: DriveInverterState
DriveInverterStateAbort: DriveInverterState
DriveInverterStateEnable: DriveInverterState
HvilStatusUnknown: HvilStatus
HvilStatusFault: HvilStatus
HvilStatusOK: HvilStatus
WindowStateUnknown: WindowState
WindowStateClosed: WindowState
WindowStatePartiallyOpen: WindowState
WindowStateOpened: WindowState
SeatFoldPositionUnknown: SeatFoldPosition
SeatFoldPositionSNA: SeatFoldPosition
SeatFoldPositionFaulted: SeatFoldPosition
SeatFoldPositionNotConfigured: SeatFoldPosition
SeatFoldPositionFolded: SeatFoldPosition
SeatFoldPositionUnfolded: SeatFoldPosition
TractorAirStatusUnknown: TractorAirStatus
TractorAirStatusNotAvailable: TractorAirStatus
TractorAirStatusError: TractorAirStatus
TractorAirStatusCharged: TractorAirStatus
TractorAirStatusBuildingPressureIntermediate: TractorAirStatus
TractorAirStatusExhaustingPressureIntermediate: TractorAirStatus
TractorAirStatusExhausted: TractorAirStatus
TrailerAirStatusUnknown: TrailerAirStatus
TrailerAirStatusSNA: TrailerAirStatus
TrailerAirStatusInvalid: TrailerAirStatus
TrailerAirStatusBobtailMode: TrailerAirStatus
TrailerAirStatusCharged: TrailerAirStatus
TrailerAirStatusBuildingPressureIntermediate: TrailerAirStatus
TrailerAirStatusExhaustingPressureIntermediate: TrailerAirStatus
TrailerAirStatusExhausted: TrailerAirStatus
HvacAutoModeStateUnknown: HvacAutoModeState
HvacAutoModeStateOn: HvacAutoModeState
HvacAutoModeStateOverride: HvacAutoModeState
CabinOverheatProtectionModeStateUnknown: CabinOverheatProtectionModeState
CabinOverheatProtectionModeStateOff: CabinOverheatProtectionModeState
CabinOverheatProtectionModeStateOn: CabinOverheatProtectionModeState
CabinOverheatProtectionModeStateFanOnly: CabinOverheatProtectionModeState
ClimateOverheatProtectionTempLimitUnknown: ClimateOverheatProtectionTempLimit
ClimateOverheatProtectionTempLimitHigh: ClimateOverheatProtectionTempLimit
ClimateOverheatProtectionTempLimitMedium: ClimateOverheatProtectionTempLimit
ClimateOverheatProtectionTempLimitLow: ClimateOverheatProtectionTempLimit
DefrostModeStateUnknown: DefrostModeState
DefrostModeStateOff: DefrostModeState
DefrostModeStateNormal: DefrostModeState
DefrostModeStateMax: DefrostModeState
DefrostModeStateAutoDefog: DefrostModeState
ClimateKeeperModeStateUnknown: ClimateKeeperModeState
ClimateKeeperModeStateOff: ClimateKeeperModeState
ClimateKeeperModeStateOn: ClimateKeeperModeState
ClimateKeeperModeStateDog: ClimateKeeperModeState
ClimateKeeperModeStateParty: ClimateKeeperModeState
HvacPowerStateUnknown: HvacPowerState
HvacPowerStateOff: HvacPowerState
HvacPowerStateOn: HvacPowerState
HvacPowerStatePrecondition: HvacPowerState
HvacPowerStateOverheatProtect: HvacPowerState
FastChargerUnknown: FastCharger
FastChargerSupercharger: FastCharger
FastChargerCHAdeMO: FastCharger
FastChargerGB: FastCharger
FastChargerACSingleWireCAN: FastCharger
FastChargerCombo: FastCharger
FastChargerMCSingleWireCAN: FastCharger
FastChargerOther: FastCharger
FastChargerSNA: FastCharger
CableTypeUnknown: CableType
CableTypeIEC: CableType
CableTypeSAE: CableType
CableTypeGB_AC: CableType
CableTypeGB_DC: CableType
CableTypeSNA: CableType
TonneauTentModeStateUnknown: TonneauTentModeState
TonneauTentModeStateInactive: TonneauTentModeState
TonneauTentModeStateMoving: TonneauTentModeState
TonneauTentModeStateFailed: TonneauTentModeState
TonneauTentModeStateActive: TonneauTentModeState
TonneauPositionStateUnknown: TonneauPositionState
TonneauPositionStateInvalid: TonneauPositionState
TonneauPositionStateClosed: TonneauPositionState
TonneauPositionStatePartiallyOpen: TonneauPositionState
TonneauPositionStateFullyOpen: TonneauPositionState
PowershareStateUnknown: PowershareState
PowershareStateInactive: PowershareState
PowershareStateHandshaking: PowershareState
PowershareStateInit: PowershareState
PowershareStateEnabled: PowershareState
PowershareStateEnabledReconnectingSoon: PowershareState
PowershareStateStopped: PowershareState
PowershareStopReasonStatusUnknown: PowershareStopReasonStatus
PowershareStopReasonStatusNone: PowershareStopReasonStatus
PowershareStopReasonStatusSOCTooLow: PowershareStopReasonStatus
PowershareStopReasonStatusRetry: PowershareStopReasonStatus
PowershareStopReasonStatusFault: PowershareStopReasonStatus
PowershareStopReasonStatusUser: PowershareStopReasonStatus
PowershareStopReasonStatusReconnecting: PowershareStopReasonStatus
PowershareStopReasonStatusAuthentication: PowershareStopReasonStatus
PowershareTypeStatusUnknown: PowershareTypeStatus
PowershareTypeStatusNone: PowershareTypeStatus
PowershareTypeStatusLoad: PowershareTypeStatus
PowershareTypeStatusHome: PowershareTypeStatus
DisplayStateUnknown: DisplayState
DisplayStateOff: DisplayState
DisplayStateDim: DisplayState
DisplayStateAccessory: DisplayState
DisplayStateOn: DisplayState
DisplayStateDriving: DisplayState
DisplayStateCharging: DisplayState
DisplayStateLock: DisplayState
DisplayStateSentry: DisplayState
DisplayStateDog: DisplayState
DisplayStateEntertainment: DisplayState
DistanceUnitUnknown: DistanceUnit
DistanceUnitMiles: DistanceUnit
DistanceUnitKilometers: DistanceUnit
TemperatureUnitUnknown: TemperatureUnit
TemperatureUnitFahrenheit: TemperatureUnit
TemperatureUnitCelsius: TemperatureUnit
PressureUnitUnknown: PressureUnit
PressureUnitPsi: PressureUnit
PressureUnitBar: PressureUnit
ChargeUnitUnknown: ChargeUnitPreference
ChargeUnitDistance: ChargeUnitPreference
ChargeUnitPercent: ChargeUnitPreference
SunroofInstalledStateUnknown: SunroofInstalledState
SunroofInstalledStateNotInstalled: SunroofInstalledState
SunroofInstalledStateGen1Installed: SunroofInstalledState
SunroofInstalledStateGen2Installed: SunroofInstalledState
TurnSignalStateUnknown: TurnSignalState
TurnSignalStateOff: TurnSignalState
TurnSignalStateLeft: TurnSignalState
TurnSignalStateRight: TurnSignalState
TurnSignalStateBoth: TurnSignalState
MediaStatusUnknown: MediaStatus
MediaStatusStopped: MediaStatus
MediaStatusPlaying: MediaStatus
MediaStatusPaused: MediaStatus

class LocationValue(_message.Message):
    __slots__ = ("latitude", "longitude")
    LATITUDE_FIELD_NUMBER: _ClassVar[int]
    LONGITUDE_FIELD_NUMBER: _ClassVar[int]
    latitude: float
    longitude: float
    def __init__(self, latitude: _Optional[float] = ..., longitude: _Optional[float] = ...) -> None: ...

class Doors(_message.Message):
    __slots__ = ("DriverFront", "DriverRear", "PassengerFront", "PassengerRear", "TrunkFront", "TrunkRear")
    DRIVERFRONT_FIELD_NUMBER: _ClassVar[int]
    DRIVERREAR_FIELD_NUMBER: _ClassVar[int]
    PASSENGERFRONT_FIELD_NUMBER: _ClassVar[int]
    PASSENGERREAR_FIELD_NUMBER: _ClassVar[int]
    TRUNKFRONT_FIELD_NUMBER: _ClassVar[int]
    TRUNKREAR_FIELD_NUMBER: _ClassVar[int]
    DriverFront: bool
    DriverRear: bool
    PassengerFront: bool
    PassengerRear: bool
    TrunkFront: bool
    TrunkRear: bool
    def __init__(self, DriverFront: bool = ..., DriverRear: bool = ..., PassengerFront: bool = ..., PassengerRear: bool = ..., TrunkFront: bool = ..., TrunkRear: bool = ...) -> None: ...

class TireLocation(_message.Message):
    __slots__ = ("front_left", "front_right", "rear_left", "rear_right", "semi_middle_axle_left_2", "semi_middle_axle_right_2", "semi_rear_axle_left", "semi_rear_axle_right", "semi_rear_axle_left_2", "semi_rear_axle_right_2")
    FRONT_LEFT_FIELD_NUMBER: _ClassVar[int]
    FRONT_RIGHT_FIELD_NUMBER: _ClassVar[int]
    REAR_LEFT_FIELD_NUMBER: _ClassVar[int]
    REAR_RIGHT_FIELD_NUMBER: _ClassVar[int]
    SEMI_MIDDLE_AXLE_LEFT_2_FIELD_NUMBER: _ClassVar[int]
    SEMI_MIDDLE_AXLE_RIGHT_2_FIELD_NUMBER: _ClassVar[int]
    SEMI_REAR_AXLE_LEFT_FIELD_NUMBER: _ClassVar[int]
    SEMI_REAR_AXLE_RIGHT_FIELD_NUMBER: _ClassVar[int]
    SEMI_REAR_AXLE_LEFT_2_FIELD_NUMBER: _ClassVar[int]
    SEMI_REAR_AXLE_RIGHT_2_FIELD_NUMBER: _ClassVar[int]
    front_left: bool
    front_right: bool
    rear_left: bool
    rear_right: bool
    semi_middle_axle_left_2: bool
    semi_middle_axle_right_2: bool
    semi_rear_axle_left: bool
    semi_rear_axle_right: bool
    semi_rear_axle_left_2: bool
    semi_rear_axle_right_2: bool
    def __init__(self, front_left: bool = ..., front_right: bool = ..., rear_left: bool = ..., rear_right: bool = ..., semi_middle_axle_left_2: bool = ..., semi_middle_axle_right_2: bool = ..., semi_rear_axle_left: bool = ..., semi_rear_axle_right: bool = ..., semi_rear_axle_left_2: bool = ..., semi_rear_axle_right_2: bool = ...) -> None: ...

class Time(_message.Message):
    __slots__ = ("hour", "minute", "second")
    HOUR_FIELD_NUMBER: _ClassVar[int]
    MINUTE_FIELD_NUMBER: _ClassVar[int]
    SECOND_FIELD_NUMBER: _ClassVar[int]
    hour: int
    minute: int
    second: int
    def __init__(self, hour: _Optional[int] = ..., minute: _Optional[int] = ..., second: _Optional[int] = ...) -> None: ...

class Value(_message.Message):
    __slots__ = ("string_value", "int_value", "long_value", "float_value", "double_value", "boolean_value", "location_value", "charging_value", "shift_state_value", "invalid", "lane_assist_level_value", "scheduled_charging_mode_value", "sentry_mode_state_value", "speed_assist_level_value", "bms_state_value", "buckle_status_value", "car_type_value", "charge_port_value", "charge_port_latch_value", "door_value", "drive_inverter_state_value", "hvil_status_value", "window_state_value", "seat_fold_position_value", "tractor_air_status_value", "follow_distance_value", "forward_collision_sensitivity_value", "guest_mode_mobile_access_value", "trailer_air_status_value", "time_value", "detailed_charge_state_value", "hvac_auto_mode_value", "cabin_overheat_protection_mode_value", "cabin_overheat_protection_temperature_limit_value", "defrost_mode_value", "climate_keeper_mode_value", "hvac_power_value", "tire_location_value", "fast_charger_value", "cable_type_value", "tonneau_tent_mode_value", "tonneau_position_value", "powershare_type_value", "powershare_state_value", "powershare_stop_reason_value", "display_state_value", "distance_unit_value", "temperature_unit_value", "pressure_unit_value", "charge_unit_preference_value", "turn_signal_state_value", "media_status_value", "sunroof_installed_state_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    LONG_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT_VALUE_FIELD_NUMBER: _ClassVar[int]
    DOUBLE_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOLEAN_VALUE_FIELD_NUMBER: _ClassVar[int]
    LOCATION_VALUE_FIELD_NUMBER: _ClassVar[int]
    CHARGING_VALUE_FIELD_NUMBER: _ClassVar[int]
    SHIFT_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    INVALID_FIELD_NUMBER: _ClassVar[int]
    LANE_ASSIST_LEVEL_VALUE_FIELD_NUMBER: _ClassVar[int]
    SCHEDULED_CHARGING_MODE_VALUE_FIELD_NUMBER: _ClassVar[int]
    SENTRY_MODE_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    SPEED_ASSIST_LEVEL_VALUE_FIELD_NUMBER: _ClassVar[int]
    BMS_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    BUCKLE_STATUS_VALUE_FIELD_NUMBER: _ClassVar[int]
    CAR_TYPE_VALUE_FIELD_NUMBER: _ClassVar[int]
    CHARGE_PORT_VALUE_FIELD_NUMBER: _ClassVar[int]
    CHARGE_PORT_LATCH_VALUE_FIELD_NUMBER: _ClassVar[int]
    DOOR_VALUE_FIELD_NUMBER: _ClassVar[int]
    DRIVE_INVERTER_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    HVIL_STATUS_VALUE_FIELD_NUMBER: _ClassVar[int]
    WINDOW_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    SEAT_FOLD_POSITION_VALUE_FIELD_NUMBER: _ClassVar[int]
    TRACTOR_AIR_STATUS_VALUE_FIELD_NUMBER: _ClassVar[int]
    FOLLOW_DISTANCE_VALUE_FIELD_NUMBER: _ClassVar[int]
    FORWARD_COLLISION_SENSITIVITY_VALUE_FIELD_NUMBER: _ClassVar[int]
    GUEST_MODE_MOBILE_ACCESS_VALUE_FIELD_NUMBER: _ClassVar[int]
    TRAILER_AIR_STATUS_VALUE_FIELD_NUMBER: _ClassVar[int]
    TIME_VALUE_FIELD_NUMBER: _ClassVar[int]
    DETAILED_CHARGE_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    HVAC_AUTO_MODE_VALUE_FIELD_NUMBER: _ClassVar[int]
    CABIN_OVERHEAT_PROTECTION_MODE_VALUE_FIELD_NUMBER: _ClassVar[int]
    CABIN_OVERHEAT_PROTECTION_TEMPERATURE_LIMIT_VALUE_FIELD_NUMBER: _ClassVar[int]
    DEFROST_MODE_VALUE_FIELD_NUMBER: _ClassVar[int]
    CLIMATE_KEEPER_MODE_VALUE_FIELD_NUMBER: _ClassVar[int]
    HVAC_POWER_VALUE_FIELD_NUMBER: _ClassVar[int]
    TIRE_LOCATION_VALUE_FIELD_NUMBER: _ClassVar[int]
    FAST_CHARGER_VALUE_FIELD_NUMBER: _ClassVar[int]
    CABLE_TYPE_VALUE_FIELD_NUMBER: _ClassVar[int]
    TONNEAU_TENT_MODE_VALUE_FIELD_NUMBER: _ClassVar[int]
    TONNEAU_POSITION_VALUE_FIELD_NUMBER: _ClassVar[int]
    POWERSHARE_TYPE_VALUE_FIELD_NUMBER: _ClassVar[int]
    POWERSHARE_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    POWERSHARE_STOP_REASON_VALUE_FIELD_NUMBER: _ClassVar[int]
    DISPLAY_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    DISTANCE_UNIT_VALUE_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_UNIT_VALUE_FIELD_NUMBER: _ClassVar[int]
    PRESSURE_UNIT_VALUE_FIELD_NUMBER: _ClassVar[int]
    CHARGE_UNIT_PREFERENCE_VALUE_FIELD_NUMBER: _ClassVar[int]
    TURN_SIGNAL_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    MEDIA_STATUS_VALUE_FIELD_NUMBER: _ClassVar[int]
    SUNROOF_INSTALLED_STATE_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    int_value: int
    long_value: int
    float_value: float
    double_value: float
    boolean_value: bool
    location_value: LocationValue
    charging_value: ChargingState
    shift_state_value: ShiftState
    invalid: bool
    lane_assist_level_value: LaneAssistLevel
    scheduled_charging_mode_value: ScheduledChargingModeValue
    sentry_mode_state_value: SentryModeState
    speed_assist_level_value: SpeedAssistLevel
    bms_state_value: BMSStateValue
    buckle_status_value: BuckleStatus
    car_type_value: CarTypeValue
    charge_port_value: ChargePortValue
    charge_port_latch_value: ChargePortLatchValue
    door_value: Doors
    drive_inverter_state_value: DriveInverterState
    hvil_status_value: HvilStatus
    window_state_value: WindowState
    seat_fold_position_value: SeatFoldPosition
    tractor_air_status_value: TractorAirStatus
    follow_distance_value: FollowDistance
    forward_collision_sensitivity_value: ForwardCollisionSensitivity
    guest_mode_mobile_access_value: GuestModeMobileAccess
    trailer_air_status_value: TrailerAirStatus
    time_value: Time
    detailed_charge_state_value: DetailedChargeStateValue
    hvac_auto_mode_value: HvacAutoModeState
    cabin_overheat_protection_mode_value: CabinOverheatProtectionModeState
    cabin_overheat_protection_temperature_limit_value: ClimateOverheatProtectionTempLimit
    defrost_mode_value: DefrostModeState
    climate_keeper_mode_value: ClimateKeeperModeState
    hvac_power_value: HvacPowerState
    tire_location_value: TireLocation
    fast_charger_value: FastCharger
    cable_type_value: CableType
    tonneau_tent_mode_value: TonneauTentModeState
    tonneau_position_value: TonneauPositionState
    powershare_type_value: PowershareTypeStatus
    powershare_state_value: PowershareState
    powershare_stop_reason_value: PowershareStopReasonStatus
    display_state_value: DisplayState
    distance_unit_value: DistanceUnit
    temperature_unit_value: TemperatureUnit
    pressure_unit_value: PressureUnit
    charge_unit_preference_value: ChargeUnitPreference
    turn_signal_state_value: TurnSignalState
    media_status_value: MediaStatus
    sunroof_installed_state_value: SunroofInstalledState
    def __init__(self, string_value: _Optional[str] = ..., int_value: _Optional[int] = ..., long_value: _Optional[int] = ..., float_value: _Optional[float] = ..., double_value: _Optional[float] = ..., boolean_value: bool = ..., location_value: _Optional[_Union[LocationValue, _Mapping]] = ..., charging_value: _Optional[_Union[ChargingState, str]] = ..., shift_state_value: _Optional[_Union[ShiftState, str]] = ..., invalid: bool = ..., lane_assist_level_value: _Optional[_Union[LaneAssistLevel, str]] = ..., scheduled_charging_mode_value: _Optional[_Union[ScheduledChargingModeValue, str]] = ..., sentry_mode_state_value: _Optional[_Union[SentryModeState, str]] = ..., speed_assist_level_value: _Optional[_Union[SpeedAssistLevel, str]] = ..., bms_state_value: _Optional[_Union[BMSStateValue, str]] = ..., buckle_status_value: _Optional[_Union[BuckleStatus, str]] = ..., car_type_value: _Optional[_Union[CarTypeValue, str]] = ..., charge_port_value: _Optional[_Union[ChargePortValue, str]] = ..., charge_port_latch_value: _Optional[_Union[ChargePortLatchValue, str]] = ..., door_value: _Optional[_Union[Doors, _Mapping]] = ..., drive_inverter_state_value: _Optional[_Union[DriveInverterState, str]] = ..., hvil_status_value: _Optional[_Union[HvilStatus, str]] = ..., window_state_value: _Optional[_Union[WindowState, str]] = ..., seat_fold_position_value: _Optional[_Union[SeatFoldPosition, str]] = ..., tractor_air_status_value: _Optional[_Union[TractorAirStatus, str]] = ..., follow_distance_value: _Optional[_Union[FollowDistance, str]] = ..., forward_collision_sensitivity_value: _Optional[_Union[ForwardCollisionSensitivity, str]] = ..., guest_mode_mobile_access_value: _Optional[_Union[GuestModeMobileAccess, str]] = ..., trailer_air_status_value: _Optional[_Union[TrailerAirStatus, str]] = ..., time_value: _Optional[_Union[Time, _Mapping]] = ..., detailed_charge_state_value: _Optional[_Union[DetailedChargeStateValue, str]] = ..., hvac_auto_mode_value: _Optional[_Union[HvacAutoModeState, str]] = ..., cabin_overheat_protection_mode_value: _Optional[_Union[CabinOverheatProtectionModeState, str]] = ..., cabin_overheat_protection_temperature_limit_value: _Optional[_Union[ClimateOverheatProtectionTempLimit, str]] = ..., defrost_mode_value: _Optional[_Union[DefrostModeState, str]] = ..., climate_keeper_mode_value: _Optional[_Union[ClimateKeeperModeState, str]] = ..., hvac_power_value: _Optional[_Union[HvacPowerState, str]] = ..., tire_location_value: _Optional[_Union[TireLocation, _Mapping]] = ..., fast_charger_value: _Optional[_Union[FastCharger, str]] = ..., cable_type_value: _Optional[_Union[CableType, str]] = ..., tonneau_tent_mode_value: _Optional[_Union[TonneauTentModeState, str]] = ..., tonneau_position_value: _Optional[_Union[TonneauPositionState, str]] = ..., powershare_type_value: _Optional[_Union[PowershareTypeStatus, str]] = ..., powershare_state_value: _Optional[_Union[PowershareState, str]] = ..., powershare_stop_reason_value: _Optional[_Union[PowershareStopReasonStatus, str]] = ..., display_state_value: _Optional[_Union[DisplayState, str]] = ..., distance_unit_value: _Optional[_Union[DistanceUnit, str]] = ..., temperature_unit_value: _Optional[_Union[TemperatureUnit, str]] = ..., pressure_unit_value: _Optional[_Union[PressureUnit, str]] = ..., charge_unit_preference_value: _Optional[_Union[ChargeUnitPreference, str]] = ..., turn_signal_state_value: _Optional[_Union[TurnSignalState, str]] = ..., media_status_value: _Optional[_Union[MediaStatus, str]] = ..., sunroof_installed_state_value: _Optional[_Union[SunroofInstalledState, str]] = ...) -> None: ...

class Datum(_message.Message):
    __slots__ = ("key", "value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: Field
    value: Value
    def __init__(self, key: _Optional[_Union[Field, str]] = ..., value: _Optional[_Union[Value, _Mapping]] = ...) -> None: ...

class Payload(_message.Message):
    __slots__ = ("data", "created_at", "vin", "is_resend")
    DATA_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    VIN_FIELD_NUMBER: _ClassVar[int]
    IS_RESEND_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedCompositeFieldContainer[Datum]
    created_at: _timestamp_pb2.Timestamp
    vin: str
    is_resend: bool
    def __init__(self, data: _Optional[_Iterable[_Union[Datum, _Mapping]]] = ..., created_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., vin: _Optional[str] = ..., is_resend: bool = ...) -> None: ...



"""Full-screen Textual TUI dashboard for telemetry + server monitoring.

Multi-panel dashboard with categorized vehicle widgets, keybindings for
vehicle commands, and a help/info modal screen.
"""

from __future__ import annotations

import asyncio
import contextlib
import logging
import os
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any, ClassVar

from textual.app import App, ComposeResult, SystemCommand
from textual.binding import Binding
from textual.containers import Grid, Horizontal, Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import DataTable, Footer, Header, RichLog, Static

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from textual.screen import Screen

    from tescmd.output.rich_output import DisplayUnits
    from tescmd.telemetry.decoder import TelemetryFrame
    from tescmd.triggers.manager import TriggerManager

logger = logging.getLogger(__name__)


def _make_cli_runner() -> Any:
    """Create a CliRunner with stderr separation.

    Click 8.2 removed the ``mix_stderr`` parameter (stderr is always
    separate).  Click 8.1 defaults to ``mix_stderr=True``, so we pass
    ``False`` when supported.  Routed through ``Any`` to avoid a
    version-dependent ``type: ignore`` that fails strict mypy on one
    version or the other.
    """
    from click.testing import CliRunner as _Runner

    _ctor: Any = _Runner
    try:
        return _ctor(mix_stderr=False)
    except TypeError:
        return _Runner()


# ---------------------------------------------------------------------------
# Panel field mapping — declarative schema for the dashboard layout
# ---------------------------------------------------------------------------

PANEL_FIELDS: dict[str, tuple[str, list[str]]] = {
    "battery": (
        "Battery & Charging",
        [
            "Soc",
            "BatteryLevel",
            "EstBatteryRange",
            "IdealBatteryRange",
            "RatedRange",
            "PackVoltage",
            "PackCurrent",
            "ChargeState",
            "DetailedChargeState",
            "ChargeLimitSoc",
            "TimeToFullCharge",
            "ACChargingPower",
            "DCChargingPower",
            "ChargerVoltage",
            "ChargeAmps",
            "ChargePortDoorOpen",
            "ChargePortLatch",
            "FastChargerPresent",
            "FastChargerType",
            "ChargingCableType",
            "BatteryHeaterOn",
            "EnergyRemaining",
            "ChargeCurrentRequest",
            "ChargeCurrentRequestMax",
            "ChargeRateMilePerHour",
            "EstimatedHoursToChargeTermination",
        ],
    ),
    "climate": (
        "Climate",
        [
            "InsideTemp",
            "OutsideTemp",
            "HvacLeftTemperatureRequest",
            "HvacRightTemperatureRequest",
            "HvacPower",
            "HvacFanStatus",
            "HvacFanSpeed",
            "HvacACEnabled",
            "HvacAutoMode",
            "DefrostMode",
            "DefrostForPreconditioning",
            "SeatHeaterLeft",
            "SeatHeaterRight",
            "SeatHeaterRearLeft",
            "SeatHeaterRearCenter",
            "SeatHeaterRearRight",
            "HvacSteeringWheelHeatLevel",
            "HvacSteeringWheelHeatAuto",
            "AutoSeatClimateLeft",
            "AutoSeatClimateRight",
            "ClimateSeatCoolingFrontLeft",
            "ClimateSeatCoolingFrontRight",
            "SeatVentEnabled",
            "RearDisplayHvacEnabled",
            "RearDefrostEnabled",
            "ClimateKeeperMode",
            "CabinOverheatProtectionMode",
            "CabinOverheatProtectionTemperatureLimit",
            "PreconditioningEnabled",
        ],
    ),
    "driving": (
        "Driving",
        [
            "VehicleSpeed",
            "Location",
            "Gear",
            "GpsHeading",
            "Odometer",
            "CruiseSetSpeed",
            "CruiseFollowDistance",
            "CurrentLimitMph",
            "SpeedLimitMode",
            "SpeedLimitWarning",
            "LateralAcceleration",
            "LongitudinalAcceleration",
            "PedalPosition",
            "BrakePedalPos",
            "LaneDepartureAvoidance",
            "ForwardCollisionWarning",
        ],
    ),
    "nav": (
        "Navigation",
        [
            "MilesToArrival",
            "MinutesToArrival",
            "RouteTrafficMinutesDelay",
            "DestinationName",
            "DestinationLocation",
            "OriginLocation",
            "RouteLine",
            "RouteLastUpdated",
            "ExpectedEnergyPercentAtTripArrival",
            "HomelinkNearby",
            "HomelinkDeviceCount",
            "LocatedAtHome",
            "LocatedAtWork",
            "LocatedAtFavorite",
        ],
    ),
    "security": (
        "Security",
        [
            "Locked",
            "SentryMode",
            "ValetModeEnabled",
            "DoorState",
            "FdWindow",
            "FpWindow",
            "RdWindow",
            "RpWindow",
            "DriverSeatOccupied",
            "DriverSeatBelt",
            "PassengerSeatBelt",
            "CenterDisplay",
            "RemoteStartEnabled",
            "GuestModeEnabled",
            "GuestModeMobileAccessState",
            "PinToDriveEnabled",
        ],
    ),
    "media": (
        "Media",
        [
            "MediaPlaybackStatus",
            "MediaPlaybackSource",
            "MediaNowPlayingTitle",
            "MediaNowPlayingArtist",
            "MediaNowPlayingAlbum",
            "MediaNowPlayingStation",
            "MediaNowPlayingDuration",
            "MediaNowPlayingElapsed",
            "MediaAudioVolume",
            "MediaAudioVolumeMax",
            "MediaAudioVolumeIncrement",
        ],
    ),
    "tires": (
        "Tires",
        [
            "TpmsPressureFl",
            "TpmsPressureFr",
            "TpmsPressureRl",
            "TpmsPressureRr",
            "TpmsLastSeenPressureTimeFl",
            "TpmsLastSeenPressureTimeFr",
            "TpmsLastSeenPressureTimeRl",
            "TpmsLastSeenPressureTimeRr",
            "TpmsHardWarnings",
            "TpmsSoftWarnings",
        ],
    ),
    "diagnostics": (
        "Diagnostics & Vehicle",
        [
            "ModuleTempMax",
            "ModuleTempMin",
            "BrickVoltageMax",
            "BrickVoltageMin",
            "NumBrickVoltageMax",
            "NumBrickVoltageMin",
            "NumModuleTempMax",
            "NumModuleTempMin",
            "BMSState",
            "DriveRail",
            "NotEnoughPowerToHeat",
            "DCDCEnable",
            "IsolationResistance",
            "Hvil",
            "Version",
            "SoftwareUpdateVersion",
            "SoftwareUpdateDownloadPercentComplete",
            "SoftwareUpdateInstallationPercentComplete",
            "CarType",
            "WheelType",
            "ServiceMode",
        ],
    ),
}

# Pre-compute field → panel_id lookup for O(1) routing.
_FIELD_TO_PANEL: dict[str, str] = {}
for _pid, (_title, _fields) in PANEL_FIELDS.items():
    for _f in _fields:
        _FIELD_TO_PANEL[_f] = _pid


# ---------------------------------------------------------------------------
# Activity sidebar — logging handler that funnels into an asyncio queue
# ---------------------------------------------------------------------------

# Logger-name prefix -> (short label, Rich color).
SOURCE_MAP: dict[str, tuple[str, str]] = {
    "tescmd.tui.commands": ("CMD", "cyan"),
    "tescmd.mcp.server": ("MCP", "magenta"),
    "tescmd.cli.serve": ("HTTP", "blue"),
    "tescmd.openclaw.gateway": ("CLAW", "green"),
    "tescmd.openclaw.bridge": ("CLAW", "green"),
    "tescmd.telemetry.server": ("TELEM", "yellow"),
    "tescmd.telemetry.cache_sink": ("CACHE", "dim"),
    "tescmd.api.client": ("API", "blue"),
    "tescmd.api.signed_command": ("SIGN", "bright_red"),
    "tescmd.protocol.session": ("PROTO", "bright_blue"),
}


class ActivityLogHandler(logging.Handler):
    """Logging handler that enqueues messages for the TUI activity sidebar."""

    def __init__(self, queue: asyncio.Queue[tuple[str, str, str]]) -> None:
        super().__init__()
        self._queue = queue

    def emit(self, record: logging.LogRecord) -> None:
        source = "LOG"
        color = "white"
        for prefix, (label, clr) in SOURCE_MAP.items():
            if record.name.startswith(prefix):
                source = label
                color = clr
                break
        with contextlib.suppress(asyncio.QueueFull):
            self._queue.put_nowait((source, color, self.format(record)))


def _mask_vin(vin: str) -> str:
    """Mask VIN for display — last 4 characters replaced with XXXX."""
    if len(vin) >= 4:
        return vin[:-4] + "XXXX"
    return vin


# ---------------------------------------------------------------------------
# Help / Info modal screen
# ---------------------------------------------------------------------------

_HELP_TEXT = (
    "KEYBINDINGS\n"
    "\n"
    "General\n"
    "  q          Quit\n"
    "  ?          Toggle this help screen\n"
    "\n"
    "Security\n"
    "  l          Lock doors\n"
    "  u          Unlock doors\n"
    "  h          Honk horn\n"
    "  f          Flash lights\n"
    "  s/S        Sentry on / off\n"
    "  r          Remote start\n"
    "  v/V        Valet on / off\n"
    "  g/G        Guest mode on / off\n"
    "\n"
    "Charging\n"
    "  c/C        Start / stop charging\n"
    "  p/P        Port open / close\n"
    "  m          Charge limit max\n"
    "  n          Charge limit standard\n"
    "\n"
    "Climate\n"
    "  a/A        Climate on / off\n"
    "  w/W        Wheel heater on / off\n"
    "\n"
    "Trunk & Windows\n"
    "  t          Trunk open/close\n"
    "  T          Frunk open\n"
    "\n"
    "Media\n"
    "  space      Play / Pause\n"
    "  \\]         Next track\n"
    "  \\[         Previous track\n"
    "  =          Volume up\n"
    "  -          Volume down\n"
    "\n"
    "Vehicle\n"
    "  ctrl+w     Wake vehicle\n"
    "\n"
    "Additional commands (no keybinding):\n"
    "  Auto-secure, Valet reset, PIN to drive,\n"
    "  Boombox, Precondition, Overheat protect,\n"
    "  Bioweapon defense, Auto wheel heater,\n"
    "  Defrost, Trunk close, Window vent/close,\n"
    "  Sunroof, Tonneau, Next/prev fav,\n"
    "  Supercharger nav, HomeLink, Software cancel,\n"
    "  Low power mode, Accessory power,\n"
    "  Charge schedule on/off\n"
)


class HelpScreen(ModalScreen[None]):
    """Modal help screen showing all keybindings and server info."""

    BINDINGS: ClassVar[list[Binding]] = [  # type: ignore[assignment]
        Binding("escape", "dismiss", "Close"),
        Binding("question_mark", "dismiss", "Close"),
    ]

    CSS = """
    HelpScreen {
        align: center middle;
    }
    #help-container {
        width: 64;
        max-height: 80%;
        background: $surface;
        border: thick $primary;
        padding: 1 2;
    }
    #help-title {
        text-align: center;
        text-style: bold;
        padding-bottom: 1;
    }
    #help-body {
        height: auto;
    }
    #help-extra {
        margin-top: 1;
        color: $text-muted;
    }
    """

    def __init__(self, log_path: str = "", server_info: str = "") -> None:
        super().__init__()
        self._log_path = log_path
        self._server_info = server_info

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="help-container"):
            yield Static("tescmd Dashboard Help", id="help-title")
            yield Static(_HELP_TEXT, id="help-body")
            extra_parts: list[str] = []
            if self._log_path:
                extra_parts.append(f"Log: {self._log_path}")
            if self._server_info:
                extra_parts.append(self._server_info)
            if extra_parts:
                yield Static("\n".join(extra_parts), id="help-extra")


# ---------------------------------------------------------------------------
# Main TUI application
# ---------------------------------------------------------------------------


class TelemetryTUI(App[None]):
    """Full-screen dashboard for tescmd telemetry and server monitoring.

    **Frame ingestion** works via a bounded :class:`asyncio.Queue`:

    - ``push_frame()`` enqueues frames from the telemetry fanout.
    - A background worker pulls frames and updates internal state.
    - A periodic timer (1 s) refreshes visible widgets from state.

    This decouples the high-frequency telemetry stream from the render
    cycle so the UI stays responsive.
    """

    TITLE = "tescmd"

    CSS = """
    #main-area {
        height: 1fr;
    }
    #panel-grid {
        width: 3fr;
        height: 1fr;
        grid-size: 2;
        grid-gutter: 0;
    }
    #sidebar {
        width: 1fr;
        min-width: 30;
        height: 1fr;
    }
    #triggers-table {
        height: auto;
        min-height: 5;
        max-height: 12;
        border: solid #e67e22;
        padding: 0;
    }
    #activity-log {
        height: 1fr;
        border: solid $accent;
        padding: 0 1;
    }
    .telemetry-panel {
        height: 1fr;
        border: solid $primary;
        padding: 0;
    }
    .telemetry-panel DataTable {
        height: 1fr;
    }
    #battery-panel { border: solid #f1c40f; }
    #climate-panel { border: solid #3498db; }
    #driving-panel { border: solid #2ecc71; }
    #security-panel { border: solid #e74c3c; }
    #tires-panel { border: solid #9b59b6; }
    #diagnostics-panel { border: solid #95a5a6; }
    #media-panel { border: solid #e67e22; }
    #nav-panel { border: solid #1abc9c; }
    #server-bar {
        height: auto;
        background: $panel;
        padding: 0 1;
        display: none;
    }
    #command-status {
        height: auto;
        background: $panel;
        padding: 0 1;
        display: none;
    }
    """

    BINDINGS: ClassVar[list[Binding]] = [  # type: ignore[assignment]
        # -- General --
        Binding("q", "quit", "Quit"),
        Binding("question_mark", "help", "Help"),
        # -- Security (shown in footer) --
        Binding("l", "cmd_lock", "Lock"),
        Binding("u", "cmd_unlock", "Unlock"),
        Binding("h", "cmd_honk", "Honk"),
        Binding("f", "cmd_flash", "Flash"),
        Binding("s", "cmd_sentry_on", "Sentry On"),
        Binding("S", "cmd_sentry_off", "Sentry Off", key_display="shift+s"),
        # -- Charging (shown in footer) --
        Binding("c", "cmd_charge_start", "Charge"),
        Binding("C", "cmd_charge_stop", "Charge Stop", key_display="shift+c"),
        # -- Security (hidden — accessible via palette / help) --
        Binding("r", "cmd_remote_start", "Remote Start", show=False),
        Binding("v", "cmd_valet_on", "Valet On", show=False),
        Binding("V", "cmd_valet_off", "Valet Off", key_display="shift+v", show=False),
        Binding("g", "cmd_guest_on", "Guest On", show=False),
        Binding("G", "cmd_guest_off", "Guest Off", key_display="shift+g", show=False),
        # -- Charging (hidden) --
        Binding("p", "cmd_port_open", "Port Open", show=False),
        Binding("P", "cmd_port_close", "Port Close", key_display="shift+p", show=False),
        Binding("m", "cmd_charge_max", "Charge Max", show=False),
        Binding("n", "cmd_charge_std", "Charge Std", show=False),
        # -- Climate (hidden) --
        Binding("a", "cmd_climate_on", "Climate On", show=False),
        Binding("A", "cmd_climate_off", "Climate Off", key_display="shift+a", show=False),
        Binding("w", "cmd_wheel_heater_on", "Wheel Heat On", show=False),
        Binding("W", "cmd_wheel_heater_off", "Wheel Heat Off", key_display="shift+w", show=False),
        # -- Trunk (hidden) --
        Binding("t", "cmd_trunk", "Trunk", show=False),
        Binding("T", "cmd_frunk", "Frunk", key_display="shift+t", show=False),
        # -- Media --
        Binding("space", "cmd_play_pause", "Play/Pause", show=False),
        Binding("right_square_bracket", "cmd_next_track", "]Next", show=False),
        Binding("left_square_bracket", "cmd_prev_track", "[Prev", show=False),
        Binding("equals_sign", "cmd_vol_up", "Vol+", show=False),
        Binding("minus", "cmd_vol_down", "Vol-", show=False),
        # -- Vehicle --
        Binding("ctrl+w", "cmd_wake", "Wake", show=False),
    ]

    def __init__(
        self,
        units: DisplayUnits | None = None,
        *,
        vin: str = "",
        telemetry_port: int | None = None,
    ) -> None:
        super().__init__()
        self._units = units
        self._vin = vin
        self._telemetry_port = telemetry_port

        # Telemetry state.
        self._state: dict[str, Any] = {}
        self._timestamps: dict[str, datetime] = {}
        self._frame_count: int = 0
        self._started_at: datetime = datetime.now(tz=UTC)
        self._connected: bool = False

        # Server info (set by callers after resources are ready).
        self._mcp_url: str = ""
        self._tunnel_url: str = ""
        self._sink_count: int = 0
        self._openclaw_status: str = ""
        self._cache_stats: str = ""
        self._log_path: str = ""

        # Bounded queue for frame ingestion.
        self._queue: asyncio.Queue[TelemetryFrame] = asyncio.Queue(maxsize=100)

        # Track which field names already have rows per panel.
        self._panel_table_fields: dict[str, set[str]] = {pid: set() for pid in PANEL_FIELDS}

        # Shutdown event — signalled when the TUI exits (e.g. user presses q).
        # serve.py can await this to know when to tear down the MCP server.
        self.shutdown_event: asyncio.Event = asyncio.Event()

        # Activity sidebar.
        self._activity_queue: asyncio.Queue[tuple[str, str, str]] = asyncio.Queue(maxsize=500)
        self._activity_handler: ActivityLogHandler | None = None
        self._last_frame_summary_count: int = 0
        self._ui_tick: int = 0
        self._original_propagate: dict[str, bool] = {}
        self._saved_root_handlers: list[logging.Handler] = []
        self._activity_file_handler: logging.FileHandler | None = None
        self._activity_log_path: str = ""

        # Trigger manager reference (set by serve.py after construction).
        self._trigger_manager: TriggerManager | None = None
        self._trigger_table_keys: set[str] = set()

        # Debug logger for command attempts — always writes to a file.
        self._cmd_logger = logging.getLogger("tescmd.tui.commands")
        self._cmd_log_handler: logging.FileHandler | None = None

    # -- Compose layout -------------------------------------------------------

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        with Horizontal(id="main-area"):
            with Grid(id="panel-grid"):
                for panel_id in PANEL_FIELDS:
                    with Vertical(id=f"{panel_id}-panel", classes="telemetry-panel"):
                        yield DataTable(
                            id=f"{panel_id}-table", cursor_type="none", zebra_stripes=True
                        )
            with Vertical(id="sidebar"):
                yield DataTable(id="triggers-table", cursor_type="none")
                yield RichLog(id="activity-log", wrap=True, markup=True)
        yield Static(id="server-bar")
        yield Static(id="command-status")
        yield Footer()

    def on_mount(self) -> None:
        """Set up DataTable columns per panel and start background processing."""
        for panel_id, (title, _fields) in PANEL_FIELDS.items():
            panel = self.query_one(f"#{panel_id}-panel", Vertical)
            panel.border_title = title

            table = self.query_one(f"#{panel_id}-table", DataTable)
            table.add_column("Field", key="field", width=24)
            table.add_column("Value", key="value", width=24)

        # Triggers table.
        triggers_table = self.query_one("#triggers-table", DataTable)
        triggers_table.border_title = "Triggers"
        triggers_table.add_column("Field", key="field", width=14)
        triggers_table.add_column("Condition", key="condition", width=10)
        triggers_table.add_column("Type", key="type", width=6)

        # Activity sidebar title.
        activity_log = self.query_one("#activity-log", RichLog)
        activity_log.border_title = "Activity"

        # Set up debug command log file.
        self._setup_command_log()

        # Attach activity handler to all monitored loggers.
        self._setup_activity_handler()

        # Periodic UI refresh.
        self.set_interval(1.0, self._update_ui)

        # Background worker to drain the frame queue.
        self.run_worker(self._process_queue, exclusive=True, thread=False)  # type: ignore[arg-type]

        # Background worker to drain the activity queue into the RichLog.
        self.run_worker(self._process_activity_queue, exclusive=False, thread=False)  # type: ignore[arg-type]

    def _setup_command_log(self) -> None:
        """Set up a file-based debug logger for command attempts."""
        from pathlib import Path

        # Place log next to the CSV log if set, otherwise default config dir.
        if self._log_path:
            log_dir = Path(self._log_path).parent
        else:
            log_dir = Path("~/.config/tescmd").expanduser()
            log_dir.mkdir(parents=True, exist_ok=True)

        log_file = log_dir / "tui-commands.log"
        handler = logging.FileHandler(log_file, encoding="utf-8")
        handler.setLevel(logging.DEBUG)
        fmt = "%(asctime)s  %(levelname)-5s  %(message)s"
        handler.setFormatter(logging.Formatter(fmt, datefmt="%Y-%m-%dT%H:%M:%S"))
        self._cmd_logger.addHandler(handler)
        self._cmd_logger.setLevel(logging.DEBUG)
        self._cmd_log_handler = handler
        self._cmd_log_path = str(log_file)
        self._cmd_logger.info("TUI started — VIN=%s", self._vin or "(none)")

    # -- Activity sidebar -------------------------------------------------------

    def _setup_activity_handler(self) -> None:
        """Attach a logging handler to all monitored loggers.

        Disables propagation so records go ONLY to the activity sidebar,
        not to the console (which would corrupt the Textual TUI).
        """
        handler = ActivityLogHandler(self._activity_queue)
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter("%(message)s"))
        self._activity_handler = handler

        # File handler — mirrors activity to a log file for post-mortem.
        from pathlib import Path

        if self._log_path:
            log_dir = Path(self._log_path).parent
        else:
            log_dir = Path("~/.config/tescmd").expanduser()
            log_dir.mkdir(parents=True, exist_ok=True)
        activity_file = log_dir / "tui-activity.log"
        file_handler = logging.FileHandler(activity_file, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s  %(levelname)-5s  [%(name)s]  %(message)s",
                datefmt="%Y-%m-%dT%H:%M:%S",
            )
        )
        self._activity_file_handler = file_handler
        self._activity_log_path = str(activity_file)

        # Attach handlers and disable console propagation for monitored loggers.
        for logger_name in SOURCE_MAP:
            log = logging.getLogger(logger_name)
            log.addHandler(handler)
            log.addHandler(file_handler)
            log.setLevel(min(log.level or logging.DEBUG, logging.DEBUG))
            self._original_propagate[logger_name] = log.propagate
            log.propagate = False

        # Suppress noisy library loggers that would corrupt the terminal.
        for name in (
            "uvicorn",
            "uvicorn.error",
            "uvicorn.access",
            "httpx",
            "httpcore",
            "websockets",
            "mcp",
        ):
            log = logging.getLogger(name)
            self._original_propagate[name] = log.propagate
            log.propagate = False

        # Remove root logger's stream handlers to prevent ANY stray
        # console output while the TUI owns the terminal.
        self._saved_root_handlers = logging.root.handlers[:]
        logging.root.handlers = [
            h
            for h in logging.root.handlers
            if not isinstance(h, logging.StreamHandler) or isinstance(h, logging.FileHandler)
        ]

    async def _process_activity_queue(self) -> None:
        """Background worker: drain activity queue into the RichLog widget."""
        while True:
            try:
                source, color, message = await asyncio.wait_for(
                    self._activity_queue.get(), timeout=0.5
                )
            except TimeoutError:
                continue

            ts = datetime.now(tz=UTC).strftime("%H:%M:%S")
            rich_log = self.query_one("#activity-log", RichLog)
            rich_log.write(f"[{color}]{ts} {source}[/{color}] {message}")

    def _maybe_log_frame_summary(self) -> None:
        """Every 5 UI ticks, log a frame summary if new frames arrived."""
        self._ui_tick += 1
        if self._ui_tick % 5 != 0:
            return
        current = self._frame_count
        delta = current - self._last_frame_summary_count
        if delta > 0:
            self._last_frame_summary_count = current
            ts = datetime.now(tz=UTC).strftime("%H:%M:%S")
            with contextlib.suppress(Exception):
                rich_log = self.query_one("#activity-log", RichLog)
                rich_log.write(f"[yellow]{ts} TELEM[/yellow] +{delta} frames ({current} total)")

    def _cleanup_activity_handler(self) -> None:
        """Remove activity handler and restore console logging."""
        handler = self._activity_handler
        if handler is None:
            return
        for logger_name in SOURCE_MAP:
            log = logging.getLogger(logger_name)
            log.removeHandler(handler)
            if self._activity_file_handler is not None:
                log.removeHandler(self._activity_file_handler)

        if self._activity_file_handler is not None:
            self._activity_file_handler.close()
            self._activity_file_handler = None

        # Restore propagation settings.
        for logger_name, propagate in self._original_propagate.items():
            logging.getLogger(logger_name).propagate = propagate
        self._original_propagate.clear()

        # Restore root logger handlers.
        logging.root.handlers = self._saved_root_handlers
        self._saved_root_handlers = []

        self._activity_handler = None

    # -- Frame ingestion (called from fanout) ---------------------------------

    async def push_frame(self, frame: TelemetryFrame) -> None:
        """Enqueue a telemetry frame for processing.

        Matches the :class:`~tescmd.telemetry.fanout.FrameFanout` callback
        signature.  If the queue is full, the frame is silently dropped
        to prevent memory growth.
        """
        with contextlib.suppress(asyncio.QueueFull):
            self._queue.put_nowait(frame)

    async def _process_queue(self) -> None:
        """Background worker: drain the queue and update state + widgets."""
        while True:
            try:
                frame = await asyncio.wait_for(self._queue.get(), timeout=0.5)
            except TimeoutError:
                continue

            self._frame_count += 1
            self._connected = True
            if frame.vin:
                self._vin = frame.vin

            for datum in frame.data:
                self._state[datum.field_name] = datum.value
                self._timestamps[datum.field_name] = frame.created_at
                self._update_panel_cell(datum.field_name, datum.value)

    # -- Server info setters (called from serve.py) ---------------------------

    def set_mcp_url(self, url: str) -> None:
        self._mcp_url = url

    def set_tunnel_url(self, url: str) -> None:
        self._tunnel_url = url

    def set_sink_count(self, n: int) -> None:
        self._sink_count = n

    def set_openclaw_status(self, connected: bool, send_count: int, event_count: int) -> None:
        status = "Connected" if connected else "Disconnected"
        self._openclaw_status = f"{status} (sent={send_count}, events={event_count})"

    def set_cache_stats(self, frames: int, fields: int, pending: int) -> None:
        self._cache_stats = f"{frames} frames, {fields} fields, {pending} pending"

    def set_log_path(self, path: Path | str) -> None:
        self._log_path = str(path)

    def set_trigger_manager(self, mgr: TriggerManager) -> None:
        self._trigger_manager = mgr

    # -- UI update (runs every 1 second) --------------------------------------

    def _update_ui(self) -> None:
        """Refresh all visible widgets from current state."""
        self._update_header()
        self._update_server_info()
        self._update_triggers()
        self._maybe_log_frame_summary()

    def _update_header(self) -> None:
        now = datetime.now(tz=UTC)
        uptime = now - self._started_at
        hours, remainder = divmod(int(uptime.total_seconds()), 3600)
        minutes, seconds = divmod(remainder, 60)

        # Title: VIN + connection status.
        vin_display = _mask_vin(self._vin) if self._vin else "(waiting)"
        status = "Connected" if self._connected else "Waiting"
        self.title = f"tescmd  {vin_display}  [{status}]"

        # Subtitle: frames + uptime (renders right-aligned beside clock).
        self.sub_title = (
            f"Frames: {self._frame_count:,}  Up: {hours:02d}:{minutes:02d}:{seconds:02d}"
        )

    def _update_panel_cell(self, field_name: str, value: Any) -> None:
        """Update a single field's cell in the appropriate panel DataTable."""
        panel_id = _FIELD_TO_PANEL.get(field_name, "diagnostics")
        display_value = _format_value(field_name, value, self._units)
        tracked = self._panel_table_fields[panel_id]

        table = self.query_one(f"#{panel_id}-table", DataTable)
        if field_name in tracked:
            table.update_cell(field_name, "value", display_value)
        else:
            table.add_row(field_name, display_value, key=field_name)
            tracked.add(field_name)

    def _update_server_info(self) -> None:
        parts: list[str] = []
        # Tunnel first — it's the public entry point.
        if self._tunnel_url:
            parts.append(f"Tunnel: {self._tunnel_url}")
        # MCP as relative path (port/path) when tunnel is present.
        if self._mcp_url:
            parts.append(f"MCP: {_relative_path(self._mcp_url, self._tunnel_url)}")
        # Telemetry WS port.
        if self._telemetry_port is not None:
            parts.append(f"WS: :{self._telemetry_port}")
        if self._sink_count:
            parts.append(f"Sinks: {self._sink_count}")
        if self._cache_stats:
            parts.append(f"Cache: {self._cache_stats}")
        if self._openclaw_status:
            parts.append(f"OpenClaw: {self._openclaw_status}")

        info_widget = self.query_one("#server-bar", Static)
        if parts:
            info_widget.update("  |  ".join(parts))
            info_widget.display = True
        else:
            info_widget.update("")
            info_widget.display = False

    def _update_triggers(self) -> None:
        """Refresh the triggers DataTable from the trigger manager."""
        if self._trigger_manager is None:
            return

        table = self.query_one("#triggers-table", DataTable)
        triggers = self._trigger_manager.list_all()
        current_ids = {t.id for t in triggers}

        # Remove rows for deleted triggers.
        removed = self._trigger_table_keys - current_ids
        for tid in removed:
            table.remove_row(tid)
        self._trigger_table_keys -= removed

        # Add or update rows.
        for t in triggers:
            cond = t.condition
            op = cond.operator.value
            if op == "changed":
                condition_text = "changed"
            elif op in ("enter", "leave"):
                condition_text = f"{op} geofence"
            else:
                symbols = {
                    "lt": "<",
                    "gt": ">",
                    "lte": "\u2264",
                    "gte": "\u2265",
                    "eq": "=",
                    "neq": "\u2260",
                }
                condition_text = f"{symbols.get(op, op)} {cond.value}"
            fire_type = "once" if t.once else "persist"

            if t.id in self._trigger_table_keys:
                table.update_cell(t.id, "condition", condition_text)
                table.update_cell(t.id, "type", fire_type)
            else:
                table.add_row(cond.field, condition_text, fire_type, key=t.id)
                self._trigger_table_keys.add(t.id)

    # -- Command execution (keybinding actions) --------------------------------

    def _show_command_status(self, text: str) -> None:
        widget = self.query_one("#command-status", Static)
        widget.update(text)
        widget.display = True

    def _hide_command_status(self) -> None:
        widget = self.query_one("#command-status", Static)
        widget.display = False

    def _run_command(self, cli_args: list[str], description: str) -> None:
        """Run a CLI command in a background thread worker."""
        self._show_command_status(f"Sending: {description}...")
        self._cmd_logger.info("%s  args=%s", description, cli_args)

        async def _invoke() -> None:
            from tescmd.cli.main import cli

            # Suppress httpx/httpcore logging during invocation — their INFO
            # lines write directly to the terminal and corrupt the TUI display.
            _noisy_loggers = ("httpx", "httpcore")
            saved_levels = {n: logging.getLogger(n).level for n in _noisy_loggers}
            for n in _noisy_loggers:
                logging.getLogger(n).setLevel(logging.WARNING)

            runner = _make_cli_runner()
            env = os.environ.copy()
            if self._vin:
                env["TESLA_VIN"] = self._vin
            try:
                result = runner.invoke(cli, ["--format", "json", "--wake", *cli_args], env=env)
            finally:
                for n in _noisy_loggers:
                    logging.getLogger(n).setLevel(saved_levels[n])

            if result.exit_code == 0:
                self._cmd_logger.info("OK  %s", description)
                self.call_from_thread(self._show_command_status, f"OK: {description}")
            else:
                stderr = (result.stderr or "").strip()
                stdout = (result.output or "").strip()
                exc_msg = str(result.exception) if result.exception else ""
                err = _extract_error_message(stderr, stdout, exc_msg)
                self._cmd_logger.error("FAIL  %s — %s", description, err)
                self.call_from_thread(self._show_command_status, f"FAIL: {description} — {err}")

            await asyncio.sleep(5)
            self.call_from_thread(self._hide_command_status)

        self.run_worker(_invoke, thread=True, exclusive=False)  # type: ignore[arg-type]

    # -- Quit (signals shutdown to serve.py) -----------------------------------

    async def action_quit(self) -> None:
        """Quit the TUI and signal the serve loop to shut down."""
        self._cmd_logger.info("QUIT user requested shutdown")
        self._cleanup_activity_handler()
        self._show_command_status("Shutting down...")
        self.shutdown_event.set()
        self.exit()

    # -- Help screen -----------------------------------------------------------

    def action_help(self) -> None:
        server_parts: list[str] = []
        if self._mcp_url:
            server_parts.append(f"MCP: {self._mcp_url}")
        if self._tunnel_url:
            server_parts.append(f"Tunnel: {self._tunnel_url}")
        log_paths = ""
        if self._log_path:
            log_paths += f"CSV log: {self._log_path}\n"
        cmd_log = getattr(self, "_cmd_log_path", "")
        if cmd_log:
            log_paths += f"Command log: {cmd_log}\n"
        if self._activity_log_path:
            log_paths += f"Activity log: {self._activity_log_path}\n"
        self.push_screen(
            HelpScreen(log_path=log_paths.rstrip(), server_info="  ".join(server_parts))
        )

    # -- Command palette (Ctrl+P) -----------------------------------------------

    def get_system_commands(self, screen: Screen[Any]) -> Iterable[SystemCommand]:
        """Populate the command palette with all vehicle commands."""
        yield from super().get_system_commands(screen)

        # --- Security ---
        yield SystemCommand("Lock doors", "Lock all doors", self.action_cmd_lock)
        yield SystemCommand("Unlock doors", "Unlock all doors", self.action_cmd_unlock)
        yield SystemCommand("Honk horn", "Honk the horn", self.action_cmd_honk)
        yield SystemCommand("Flash lights", "Flash headlights", self.action_cmd_flash)
        yield SystemCommand("Sentry mode on", "Enable sentry mode", self.action_cmd_sentry_on)
        yield SystemCommand("Sentry mode off", "Disable sentry mode", self.action_cmd_sentry_off)
        yield SystemCommand("Remote start", "Enable keyless driving", self.action_cmd_remote_start)
        yield SystemCommand("Valet mode on", "Enable valet mode", self.action_cmd_valet_on)
        yield SystemCommand("Valet mode off", "Disable valet mode", self.action_cmd_valet_off)
        yield SystemCommand("Valet PIN reset", "Reset valet mode PIN", self.action_cmd_valet_reset)
        yield SystemCommand(
            "Auto-secure", "Auto-lock and close windows", self.action_cmd_auto_secure
        )
        yield SystemCommand(
            "PIN to Drive on", "Enable PIN to Drive", self.action_cmd_pin_to_drive_on
        )
        yield SystemCommand(
            "PIN to Drive off", "Disable PIN to Drive", self.action_cmd_pin_to_drive_off
        )
        yield SystemCommand("Guest mode on", "Enable guest mode", self.action_cmd_guest_on)
        yield SystemCommand("Guest mode off", "Disable guest mode", self.action_cmd_guest_off)
        yield SystemCommand(
            "Boombox", "Play boombox sound", self.action_cmd_boombox, discover=False
        )
        yield SystemCommand(
            "Speed limit clear (admin)",
            "Admin clear speed limit",
            self.action_cmd_speed_clear_admin,
            discover=False,
        )

        # --- Charging ---
        yield SystemCommand(
            "Start charging", "Begin charging session", self.action_cmd_charge_start
        )
        yield SystemCommand("Stop charging", "Stop charging session", self.action_cmd_charge_stop)
        yield SystemCommand(
            "Open charge port", "Open the charge port door", self.action_cmd_port_open
        )
        yield SystemCommand(
            "Close charge port", "Close the charge port door", self.action_cmd_port_close
        )
        yield SystemCommand(
            "Charge limit: max range",
            "Set charge limit to 100%",
            self.action_cmd_charge_max,
        )
        yield SystemCommand(
            "Charge limit: standard",
            "Set charge limit to standard (80%)",
            self.action_cmd_charge_std,
        )
        yield SystemCommand(
            "Charge limit: 50%",
            "Set charge limit to 50%",
            lambda: self._run_command(["charge", "limit", "50"], "Charge limit 50%"),
        )
        yield SystemCommand(
            "Charge limit: 60%",
            "Set charge limit to 60%",
            lambda: self._run_command(["charge", "limit", "60"], "Charge limit 60%"),
        )
        yield SystemCommand(
            "Charge limit: 70%",
            "Set charge limit to 70%",
            lambda: self._run_command(["charge", "limit", "70"], "Charge limit 70%"),
        )
        yield SystemCommand(
            "Charge limit: 80%",
            "Set charge limit to 80%",
            lambda: self._run_command(["charge", "limit", "80"], "Charge limit 80%"),
        )
        yield SystemCommand(
            "Charge limit: 90%",
            "Set charge limit to 90%",
            lambda: self._run_command(["charge", "limit", "90"], "Charge limit 90%"),
        )
        yield SystemCommand(
            "Charge amps: 8A",
            "Set charge amps to 8",
            lambda: self._run_command(["charge", "amps", "8"], "Charge amps 8A"),
        )
        yield SystemCommand(
            "Charge amps: 16A",
            "Set charge amps to 16",
            lambda: self._run_command(["charge", "amps", "16"], "Charge amps 16A"),
        )
        yield SystemCommand(
            "Charge amps: 24A",
            "Set charge amps to 24",
            lambda: self._run_command(["charge", "amps", "24"], "Charge amps 24A"),
        )
        yield SystemCommand(
            "Charge amps: 32A",
            "Set charge amps to 32",
            lambda: self._run_command(["charge", "amps", "32"], "Charge amps 32A"),
        )
        yield SystemCommand(
            "Charge amps: 48A",
            "Set charge amps to 48",
            lambda: self._run_command(["charge", "amps", "48"], "Charge amps 48A"),
        )
        yield SystemCommand(
            "Scheduled charging on",
            "Enable scheduled charging",
            self.action_cmd_charge_schedule_on,
        )
        yield SystemCommand(
            "Scheduled charging off",
            "Disable scheduled charging",
            self.action_cmd_charge_schedule_off,
        )

        # --- Climate ---
        yield SystemCommand("Climate on", "Turn on climate control", self.action_cmd_climate_on)
        yield SystemCommand("Climate off", "Turn off climate control", self.action_cmd_climate_off)
        yield SystemCommand(
            "Steering wheel heater on",
            "Enable steering wheel heater",
            self.action_cmd_wheel_heater_on,
        )
        yield SystemCommand(
            "Steering wheel heater off",
            "Disable steering wheel heater",
            self.action_cmd_wheel_heater_off,
        )
        yield SystemCommand(
            "Preconditioning on",
            "Enable battery preconditioning",
            self.action_cmd_precondition_on,
        )
        yield SystemCommand(
            "Preconditioning off",
            "Disable battery preconditioning",
            self.action_cmd_precondition_off,
        )
        yield SystemCommand(
            "Cabin overheat protection on",
            "Enable cabin overheat protection",
            self.action_cmd_overheat_on,
        )
        yield SystemCommand(
            "Cabin overheat protection off",
            "Disable cabin overheat protection",
            self.action_cmd_overheat_off,
        )
        yield SystemCommand(
            "Bioweapon defense on",
            "Enable bioweapon defense mode",
            self.action_cmd_bioweapon_on,
        )
        yield SystemCommand(
            "Bioweapon defense off",
            "Disable bioweapon defense mode",
            self.action_cmd_bioweapon_off,
        )
        yield SystemCommand(
            "Auto steering wheel heater on",
            "Enable auto steering wheel heater",
            self.action_cmd_auto_wheel_on,
        )
        yield SystemCommand(
            "Auto steering wheel heater off",
            "Disable auto steering wheel heater",
            self.action_cmd_auto_wheel_off,
        )
        yield SystemCommand("Defrost on", "Enable max defrost", self.action_cmd_defrost_on)
        yield SystemCommand("Defrost off", "Disable max defrost", self.action_cmd_defrost_off)
        yield SystemCommand(
            "Rear defrost on",
            "Enable rear window defrost",
            lambda: self._run_command(["climate", "set", "--rear-defrost-on"], "Rear defrost on"),
        )
        yield SystemCommand(
            "Rear defrost off",
            "Disable rear window defrost",
            lambda: self._run_command(
                ["climate", "set", "--rear-defrost-off"], "Rear defrost off"
            ),
        )
        yield SystemCommand(
            "Set cabin temp 68°F / 20°C",
            "Set driver and passenger temp to 68°F",
            lambda: self._run_command(
                ["climate", "set", "--driver-temp", "20", "--passenger-temp", "20"],
                "Cabin temp 20°C",
            ),
        )
        yield SystemCommand(
            "Set cabin temp 72°F / 22°C",
            "Set driver and passenger temp to 72°F",
            lambda: self._run_command(
                ["climate", "set", "--driver-temp", "22", "--passenger-temp", "22"],
                "Cabin temp 22°C",
            ),
        )
        yield SystemCommand(
            "Set cabin temp 75°F / 24°C",
            "Set driver and passenger temp to 75°F",
            lambda: self._run_command(
                ["climate", "set", "--driver-temp", "24", "--passenger-temp", "24"],
                "Cabin temp 24°C",
            ),
        )
        yield SystemCommand(
            "Seat heater: driver high",
            "Set driver seat heater to high",
            lambda: self._run_command(
                ["climate", "seat", "driver", "3"], "Driver seat heater high"
            ),
        )
        yield SystemCommand(
            "Seat heater: driver off",
            "Turn off driver seat heater",
            lambda: self._run_command(
                ["climate", "seat", "driver", "0"], "Driver seat heater off"
            ),
        )
        yield SystemCommand(
            "Seat heater: passenger high",
            "Set passenger seat heater to high",
            lambda: self._run_command(
                ["climate", "seat", "passenger", "3"],
                "Passenger seat heater high",
            ),
        )
        yield SystemCommand(
            "Seat heater: passenger off",
            "Turn off passenger seat heater",
            lambda: self._run_command(
                ["climate", "seat", "passenger", "0"], "Passenger seat heater off"
            ),
        )
        yield SystemCommand(
            "Auto seat climate: driver on",
            "Enable driver auto seat climate",
            lambda: self._run_command(
                ["climate", "auto-seat", "driver", "--on"], "Auto seat driver on"
            ),
        )
        yield SystemCommand(
            "Auto seat climate: driver off",
            "Disable driver auto seat climate",
            lambda: self._run_command(
                ["climate", "auto-seat", "driver", "--off"], "Auto seat driver off"
            ),
        )
        yield SystemCommand(
            "Climate keeper: Dog mode",
            "Set climate keeper to Dog mode",
            lambda: self._run_command(["climate", "keeper", "dog"], "Climate keeper: Dog"),
        )
        yield SystemCommand(
            "Climate keeper: Camp mode",
            "Set climate keeper to Camp mode",
            lambda: self._run_command(["climate", "keeper", "camp"], "Climate keeper: Camp"),
        )
        yield SystemCommand(
            "Climate keeper: off",
            "Turn off climate keeper mode",
            lambda: self._run_command(["climate", "keeper", "off"], "Climate keeper: off"),
        )
        yield SystemCommand(
            "Cabin overheat temp: low",
            "Set overheat protection temp to low (90°F/32°C)",
            lambda: self._run_command(["climate", "cop-temp", "low"], "Overheat temp: low"),
        )
        yield SystemCommand(
            "Cabin overheat temp: medium",
            "Set overheat protection temp to medium (100°F/38°C)",
            lambda: self._run_command(["climate", "cop-temp", "medium"], "Overheat temp: medium"),
        )
        yield SystemCommand(
            "Cabin overheat temp: high",
            "Set overheat protection temp to high (110°F/43°C)",
            lambda: self._run_command(["climate", "cop-temp", "high"], "Overheat temp: high"),
        )

        # --- Trunk & Windows ---
        yield SystemCommand("Open trunk", "Open/close rear trunk", self.action_cmd_trunk)
        yield SystemCommand("Close trunk", "Close rear trunk", self.action_cmd_trunk_close)
        yield SystemCommand("Open frunk", "Open front trunk", self.action_cmd_frunk)
        yield SystemCommand("Vent windows", "Vent all windows", self.action_cmd_window_vent)
        yield SystemCommand("Close windows", "Close all windows", self.action_cmd_window_close)
        yield SystemCommand("Vent sunroof", "Vent the sunroof", self.action_cmd_sunroof_vent)
        yield SystemCommand("Close sunroof", "Close the sunroof", self.action_cmd_sunroof_close)
        yield SystemCommand("Open tonneau", "Open the tonneau cover", self.action_cmd_tonneau_open)
        yield SystemCommand(
            "Close tonneau", "Close the tonneau cover", self.action_cmd_tonneau_close
        )
        yield SystemCommand("Stop tonneau", "Stop tonneau movement", self.action_cmd_tonneau_stop)

        # --- Media ---
        yield SystemCommand("Play / Pause", "Toggle media playback", self.action_cmd_play_pause)
        yield SystemCommand("Next track", "Skip to next track", self.action_cmd_next_track)
        yield SystemCommand("Previous track", "Go to previous track", self.action_cmd_prev_track)
        yield SystemCommand("Next favorite", "Skip to next favorite", self.action_cmd_next_fav)
        yield SystemCommand(
            "Previous favorite", "Go to previous favorite", self.action_cmd_prev_fav
        )
        yield SystemCommand("Volume up", "Increase volume", self.action_cmd_vol_up)
        yield SystemCommand("Volume down", "Decrease volume", self.action_cmd_vol_down)
        yield SystemCommand(
            "Volume: mute (0)",
            "Set volume to 0",
            lambda: self._run_command(["media", "adjust-volume", "0"], "Volume 0"),
        )
        yield SystemCommand(
            "Volume: 25%",
            "Set volume to 25%",
            lambda: self._run_command(["media", "adjust-volume", "2.75"], "Volume 25%"),
        )
        yield SystemCommand(
            "Volume: 50%",
            "Set volume to 50%",
            lambda: self._run_command(["media", "adjust-volume", "5.5"], "Volume 50%"),
        )

        # --- Navigation ---
        yield SystemCommand(
            "Navigate to Supercharger",
            "Navigate to nearest Supercharger",
            self.action_cmd_nav_supercharger,
        )
        yield SystemCommand(
            "Trigger HomeLink",
            "Trigger HomeLink (garage door)",
            self.action_cmd_nav_homelink,
        )
        yield SystemCommand(
            "Navigation: get GPS position",
            "Read current vehicle GPS coordinates",
            lambda: self._run_command(["nav", "gps"], "GPS position"),
        )

        # --- Software ---
        yield SystemCommand(
            "Cancel software update",
            "Cancel a pending software update",
            self.action_cmd_software_cancel,
        )
        yield SystemCommand(
            "Schedule software update now",
            "Schedule software update to install in 60s",
            lambda: self._run_command(["software", "schedule", "60"], "Schedule software update"),
        )

        # --- Vehicle ---
        yield SystemCommand("Wake vehicle", "Wake the vehicle from sleep", self.action_cmd_wake)
        yield SystemCommand(
            "Low power mode on",
            "Enable low power consumption mode",
            self.action_cmd_low_power_on,
        )
        yield SystemCommand(
            "Low power mode off",
            "Disable low power consumption mode",
            self.action_cmd_low_power_off,
        )
        yield SystemCommand(
            "Accessory power on",
            "Enable accessory power mode",
            self.action_cmd_accessory_power_on,
        )
        yield SystemCommand(
            "Accessory power off",
            "Disable accessory power mode",
            self.action_cmd_accessory_power_off,
        )
        yield SystemCommand(
            "Mobile access on",
            "Enable mobile app access",
            lambda: self._run_command(["vehicle", "mobile-access", "--on"], "Mobile access on"),
        )
        yield SystemCommand(
            "Mobile access off",
            "Disable mobile app access",
            lambda: self._run_command(["vehicle", "mobile-access", "--off"], "Mobile access off"),
        )

        # --- Sharing ---
        yield SystemCommand(
            "List driver invites",
            "List all sharing invitations",
            lambda: self._run_command(["sharing", "list-invites"], "List invites"),
        )

    # -- Security actions ------------------------------------------------------

    def action_cmd_lock(self) -> None:
        self._run_command(["security", "lock"], "Lock doors")

    def action_cmd_unlock(self) -> None:
        self._run_command(["security", "unlock"], "Unlock doors")

    def action_cmd_honk(self) -> None:
        self._run_command(["security", "honk"], "Honk horn")

    def action_cmd_flash(self) -> None:
        self._run_command(["security", "flash"], "Flash lights")

    def action_cmd_sentry_on(self) -> None:
        self._run_command(["security", "sentry", "--on"], "Sentry mode on")

    def action_cmd_sentry_off(self) -> None:
        self._run_command(["security", "sentry", "--off"], "Sentry mode off")

    def action_cmd_remote_start(self) -> None:
        self._run_command(["security", "remote-start"], "Remote start")

    def action_cmd_valet_on(self) -> None:
        self._run_command(["security", "valet", "--on"], "Valet mode on")

    def action_cmd_valet_off(self) -> None:
        self._run_command(["security", "valet", "--off"], "Valet mode off")

    def action_cmd_valet_reset(self) -> None:
        self._run_command(["security", "valet-reset"], "Valet pin reset")

    def action_cmd_auto_secure(self) -> None:
        self._run_command(["security", "auto-secure"], "Auto secure")

    def action_cmd_pin_to_drive_on(self) -> None:
        self._run_command(["security", "pin-to-drive", "--on"], "PIN to drive on")

    def action_cmd_pin_to_drive_off(self) -> None:
        self._run_command(["security", "pin-to-drive", "--off"], "PIN to drive off")

    def action_cmd_guest_on(self) -> None:
        self._run_command(["security", "guest-mode", "--on"], "Guest mode on")

    def action_cmd_guest_off(self) -> None:
        self._run_command(["security", "guest-mode", "--off"], "Guest mode off")

    def action_cmd_boombox(self) -> None:
        self._run_command(["security", "boombox"], "Boombox")

    def action_cmd_speed_clear_admin(self) -> None:
        self._run_command(["security", "speed-clear-admin"], "Speed limit clear (admin)")

    # -- Charging actions ------------------------------------------------------

    def action_cmd_charge_start(self) -> None:
        self._run_command(["charge", "start"], "Start charging")

    def action_cmd_charge_stop(self) -> None:
        self._run_command(["charge", "stop"], "Stop charging")

    def action_cmd_port_open(self) -> None:
        self._run_command(["charge", "port-open"], "Open charge port")

    def action_cmd_port_close(self) -> None:
        self._run_command(["charge", "port-close"], "Close charge port")

    def action_cmd_charge_max(self) -> None:
        self._run_command(["charge", "limit-max"], "Charge limit max")

    def action_cmd_charge_std(self) -> None:
        self._run_command(["charge", "limit-std"], "Charge limit standard")

    def action_cmd_charge_schedule_on(self) -> None:
        self._run_command(["charge", "schedule", "--enable"], "Charge schedule on")

    def action_cmd_charge_schedule_off(self) -> None:
        self._run_command(["charge", "schedule", "--disable"], "Charge schedule off")

    # -- Climate actions -------------------------------------------------------

    def action_cmd_climate_on(self) -> None:
        self._run_command(["climate", "on"], "Climate on")

    def action_cmd_climate_off(self) -> None:
        self._run_command(["climate", "off"], "Climate off")

    def action_cmd_wheel_heater_on(self) -> None:
        self._run_command(["climate", "wheel-heater", "--on"], "Wheel heater on")

    def action_cmd_wheel_heater_off(self) -> None:
        self._run_command(["climate", "wheel-heater", "--off"], "Wheel heater off")

    def action_cmd_precondition_on(self) -> None:
        self._run_command(["climate", "precondition", "--on"], "Precondition on")

    def action_cmd_precondition_off(self) -> None:
        self._run_command(["climate", "precondition", "--off"], "Precondition off")

    def action_cmd_overheat_on(self) -> None:
        self._run_command(["climate", "overheat", "--on"], "Overheat protection on")

    def action_cmd_overheat_off(self) -> None:
        self._run_command(["climate", "overheat", "--off"], "Overheat protection off")

    def action_cmd_bioweapon_on(self) -> None:
        self._run_command(["climate", "bioweapon", "--on"], "Bioweapon defense on")

    def action_cmd_bioweapon_off(self) -> None:
        self._run_command(["climate", "bioweapon", "--off"], "Bioweapon defense off")

    def action_cmd_auto_wheel_on(self) -> None:
        self._run_command(["climate", "auto-wheel", "--on"], "Auto wheel heater on")

    def action_cmd_auto_wheel_off(self) -> None:
        self._run_command(["climate", "auto-wheel", "--off"], "Auto wheel heater off")

    def action_cmd_defrost_on(self) -> None:
        self._run_command(["climate", "set", "--defrost-on"], "Defrost on")

    def action_cmd_defrost_off(self) -> None:
        self._run_command(["climate", "set", "--defrost-off"], "Defrost off")

    # -- Trunk actions ---------------------------------------------------------

    def action_cmd_trunk(self) -> None:
        self._run_command(["trunk", "open"], "Trunk open/close")

    def action_cmd_trunk_close(self) -> None:
        self._run_command(["trunk", "close"], "Trunk close")

    def action_cmd_frunk(self) -> None:
        self._run_command(["trunk", "frunk"], "Frunk open")

    def action_cmd_window_vent(self) -> None:
        self._run_command(["trunk", "window", "--vent"], "Windows vent")

    def action_cmd_window_close(self) -> None:
        self._run_command(["trunk", "window", "--close"], "Windows close")

    def action_cmd_sunroof_vent(self) -> None:
        self._run_command(["trunk", "sunroof", "--state", "vent"], "Sunroof vent")

    def action_cmd_sunroof_close(self) -> None:
        self._run_command(["trunk", "sunroof", "--state", "close"], "Sunroof close")

    def action_cmd_tonneau_open(self) -> None:
        self._run_command(["trunk", "tonneau-open"], "Tonneau open")

    def action_cmd_tonneau_close(self) -> None:
        self._run_command(["trunk", "tonneau-close"], "Tonneau close")

    def action_cmd_tonneau_stop(self) -> None:
        self._run_command(["trunk", "tonneau-stop"], "Tonneau stop")

    # -- Media actions ---------------------------------------------------------

    def action_cmd_play_pause(self) -> None:
        self._run_command(["media", "play-pause"], "Play/Pause")

    def action_cmd_next_track(self) -> None:
        self._run_command(["media", "next-track"], "Next track")

    def action_cmd_prev_track(self) -> None:
        self._run_command(["media", "prev-track"], "Previous track")

    def action_cmd_next_fav(self) -> None:
        self._run_command(["media", "next-fav"], "Next favorite")

    def action_cmd_prev_fav(self) -> None:
        self._run_command(["media", "prev-fav"], "Previous favorite")

    def action_cmd_vol_up(self) -> None:
        self._run_command(["media", "volume-up"], "Volume up")

    def action_cmd_vol_down(self) -> None:
        self._run_command(["media", "volume-down"], "Volume down")

    # -- Navigation actions ----------------------------------------------------

    def action_cmd_nav_supercharger(self) -> None:
        self._run_command(["nav", "supercharger"], "Navigate to Supercharger")

    def action_cmd_nav_homelink(self) -> None:
        self._run_command(["nav", "homelink"], "HomeLink trigger")

    # -- Software actions ------------------------------------------------------

    def action_cmd_software_cancel(self) -> None:
        self._run_command(["software", "cancel"], "Cancel software update")

    # -- Vehicle actions -------------------------------------------------------

    def action_cmd_wake(self) -> None:
        self._run_command(["vehicle", "wake", "--wait"], "Wake vehicle")

    def action_cmd_low_power_on(self) -> None:
        self._run_command(["vehicle", "low-power", "--on"], "Low power mode on")

    def action_cmd_low_power_off(self) -> None:
        self._run_command(["vehicle", "low-power", "--off"], "Low power mode off")

    def action_cmd_accessory_power_on(self) -> None:
        self._run_command(["vehicle", "accessory-power", "--on"], "Accessory power on")

    def action_cmd_accessory_power_off(self) -> None:
        self._run_command(["vehicle", "accessory-power", "--off"], "Accessory power off")


# ---------------------------------------------------------------------------
# Value formatting (ported from dashboard.py)
# ---------------------------------------------------------------------------


def _relative_path(url: str, tunnel_url: str) -> str:
    """Reduce a full URL to a relative port/path when a tunnel is active.

    If the URL shares the same scheme+host as *tunnel_url*, show only the
    path portion.  Otherwise fall back to ``:port/path`` for localhost URLs,
    or the full URL as-is.
    """
    from urllib.parse import urlparse

    parsed = urlparse(url)
    if tunnel_url:
        tunnel_parsed = urlparse(tunnel_url)
        if parsed.hostname == tunnel_parsed.hostname:
            return parsed.path or "/"
    # Localhost — show :port/path.
    if parsed.hostname in ("127.0.0.1", "localhost", "::1"):
        return f":{parsed.port}{parsed.path}" if parsed.port else parsed.path or "/"
    return url


def _extract_error_message(stderr: str, stdout: str, exc_msg: str = "") -> str:
    """Extract a human-readable error from CLI output.

    When the TUI runs commands via CliRunner, exceptions are caught by Click
    and stored in ``result.exception`` rather than being formatted by
    ``main()``'s error handler.  The *exc_msg* parameter (from
    ``str(result.exception)``) is the most reliable error source.

    Falls back to parsing JSON error envelopes from stderr, then the last
    non-httpx line, then raw stderr/stdout.
    """
    import json

    # Best source: the exception message from Click's runner.
    if exc_msg:
        return exc_msg[:120]

    # Try each line of stderr for a JSON error envelope.
    for line in stderr.splitlines():
        line = line.strip()
        if line.startswith("{"):
            try:
                data = json.loads(line)
                err = data.get("error", {})
                msg = err.get("message", "") if isinstance(err, dict) else str(err)
                if msg:
                    return msg[:120]
            except json.JSONDecodeError:
                continue

    # Fall back: last non-empty line of stderr, filtering out httpx INFO lines.
    for line in reversed(stderr.splitlines()):
        line = line.strip()
        if line and "HTTP Request:" not in line and "HTTP/1.1" not in line:
            return line[:120]

    return (stderr or stdout)[:80]


def _format_value(
    field_name: str,
    value: Any,
    units: DisplayUnits | None,
) -> str:
    """Format a telemetry value with optional unit conversion."""
    if value is None:
        return "—"

    if isinstance(value, dict):
        lat = value.get("latitude", 0.0)
        lng = value.get("longitude", 0.0)
        return f"{lat:.6f}, {lng:.6f}"

    if isinstance(value, bool):
        return "Yes" if value else "No"

    if units is not None:
        return _format_with_units(field_name, value, units)

    # No unit preferences — raw value.
    if isinstance(value, float):
        return f"{value:.2f}"
    return str(value)


_TEMP_FIELDS = frozenset(
    {
        "InsideTemp",
        "OutsideTemp",
        "HvacLeftTemperatureRequest",
        "HvacRightTemperatureRequest",
        "ModuleTempMax",
        "ModuleTempMin",
    }
)
_DISTANCE_FIELDS = frozenset(
    {
        "Odometer",
        "EstBatteryRange",
        "IdealBatteryRange",
        "RatedRange",
        "MilesToArrival",
    }
)
_SPEED_FIELDS = frozenset({"VehicleSpeed", "CruiseSetSpeed", "CurrentLimitMph"})
_PRESSURE_FIELDS = frozenset(
    {
        "TpmsPressureFl",
        "TpmsPressureFr",
        "TpmsPressureRl",
        "TpmsPressureRr",
    }
)
_PCT_FIELDS = frozenset({"Soc", "BatteryLevel", "ChargeLimitSoc"})


def _format_with_units(field_name: str, value: Any, units: DisplayUnits) -> str:
    """Apply unit conversion based on user preferences."""
    from tescmd.output.rich_output import DistanceUnit, PressureUnit, TempUnit

    if field_name in _TEMP_FIELDS and isinstance(value, (int, float)):
        if units.temp == TempUnit.F:
            return f"{value * 9 / 5 + 32:.1f}\u00b0F"
        return f"{value:.1f}\u00b0C"

    if field_name in _DISTANCE_FIELDS and isinstance(value, (int, float)):
        if units.distance == DistanceUnit.KM:
            return f"{value * 1.60934:.1f} km"
        return f"{value:.1f} mi"

    if field_name in _SPEED_FIELDS and isinstance(value, (int, float)):
        if units.distance == DistanceUnit.KM:
            return f"{value * 1.60934:.0f} km/h"
        return f"{value:.0f} mph"

    if field_name in _PRESSURE_FIELDS and isinstance(value, (int, float)):
        if units.pressure == PressureUnit.PSI:
            return f"{value * 14.5038:.1f} psi"
        return f"{value:.2f} bar"

    if field_name in _PCT_FIELDS and isinstance(value, (int, float)):
        return f"{value}%"

    # Voltage / current.
    if "Voltage" in field_name and isinstance(value, (int, float)):
        return f"{value:.1f} V"
    if ("Current" in field_name or "Amps" in field_name) and isinstance(value, (int, float)):
        return f"{value:.1f} A"

    # Power.
    if "Power" in field_name and isinstance(value, (int, float)):
        return f"{value:.2f} kW"

    # Time-to-full.
    if field_name == "TimeToFullCharge" and isinstance(value, (int, float)):
        hours = int(value)
        mins = int((value - hours) * 60)
        return f"{hours}h {mins}m" if hours else f"{mins}m"

    if isinstance(value, float):
        return f"{value:.2f}"
    return str(value)
